import pysam
import numpy as np
import sys
import logging
import argparse
from collections import defaultdict


def out_counts_bulk(f, counts, discard, stats):
    discard_keys = sorted(discard.keys())
    f.write("name\tL\tcount\tdropped\t" + "\t".join(discard_keys) + "\n")
    all_keys = stats.keys()
    all_counts = dict([(k, counts[k]) for k in all_keys])
    for mir, count in sorted(all_counts.items(), key=lambda x: -x[1]):
        # L =
        # print(f"{mir}\t{sam.get_reference_length(mir)}\t{count}\t{rev[mir]}")
        f.write(
            f"{mir}\t-1\t{count}\t{stats[mir]['n_dropped']}\t"
            + "\t".join([str(discard[reason][mir]) for reason in discard_keys])
            + "\n"
        )


class AlignmentClassifier:
    def __init__(
        self,
        sample_name,
        bam_in,
        gene_assign_mode="chrom",
        chrom_to_gene={},
        ignore_gf=["INTRONIC", "INTERGENIC"],
    ):
        self.logger = logging.getLogger(f"AlignmentClassifier({sample_name})")
        self.sample_name = sample_name
        self.bam = pysam.AlignmentFile(bam_in, check_sq=False)
        self.gene_names = {}
        self.chrom_to_gene = chrom_to_gene
        self.ignore_gf = set(ignore_gf)
        self.get_gene = {
            "chrom": self.fast_refname,
            "gn_tag": self.parse_gn_gf,
        }[gene_assign_mode]

        self.uniq_reads = set()

    def fast_refname(self, aln):
        if not aln.tid in self.gene_names:
            chrom = self.bam.get_reference_name(aln.tid)
            gene = self.chrom_to_gene.get(chrom, chrom)
            self.gene_names[aln.tid] = gene

        return self.gene_names[aln.tid]

    def parse_gn_gf(self, aln):
        if not aln.has_tag("gf"):
            return "NA"

        genes = set()
        for gf, gn in zip(aln.get_tag("gf").split(","), aln.get_tag("gn").split(",")):
            if gf not in self.ignore_gf:
                genes.add(gn)

        genes = list(genes)
        if len(genes) == 1:
            return genes[0]
        else:
            return "multi_mapper"

    def is_uniq(self, aln):
        key = (aln.query_sequence, aln.get_tag("CB"), aln.get_tag("MI"))
        u = key in self.uniq_reads
        self.uniq_reads.add(u)
        return u

    def __iter__(self):
        for aln in self.bam.fetch(until_eof=True):
            yield ClassifiedAlignment(
                self, aln, self.get_gene(aln), is_dup=not self.is_uniq(aln)
            )


class ClassifiedAlignment:
    def __init__(self, parent, aln, gene, is_dup=False):
        self.aln = aln
        self.parent = parent
        self.tags = dict(aln.get_tags())
        self.flags = set()
        self.cell = self.tags.get("CB", "NA")
        self.umi = self.tags.get("MI", "NA")
        self.gene = gene
        self.is_dup = is_dup

    def check_qname(self, keywords=[]):
        for n in keywords:
            if n in self.aln.qname:
                self.flags.add(f"name_has_{n}")
            else:
                self.flags.add(f"name_missing_{n}")

        return self

    def check_tags(self, **kw):
        for tag, keywords in kw.items():
            tagval = self.tags.get(tag, "")
            for n in keywords:
                if n in tagval:
                    self.flags.add(f"{tag}_has_{n}")
                else:
                    self.flags.add(f"{tag}_missing_{n}")

        return self

    def check_CIGAR(self):
        n_match = 0
        c5 = 0
        c3 = 0
        cigar = self.aln.cigartuples
        if cigar:
            op, n = cigar[0]
            if op == 4:  # soft-clip
                c5 = n
            elif op == 0:  # match
                n_match = max(n_match, n)

        if len(cigar) > 1:
            for op, n in cigar[1:]:
                if op == 4:  # soft-clip again, must be 3' end
                    c3 = n
                elif op == 0:  # match
                    n_match = max(n_match, n)

        self.clip5 = c5
        self.clip3 = c3
        self.n_match = n_match

        return self


class DGE:
    def __init__(self, _assert=False, main_channel="count"):
        self.main_channel = main_channel
        self.DGE_umis = defaultdict(lambda: defaultdict(set))
        self.DGE_reads = defaultdict(lambda: defaultdict(int))
        self.DGE_cells = set()
        self.DGE_genes = set()

        # marginal counters for each cell
        self.DGE_cell_reads = defaultdict(int)
        self.DGE_cell_umis = defaultdict(set)
        self.DGE_cell_genes = defaultdict(set)

        # marginal counts for each gene
        self.DGE_gene_reads = defaultdict(int)
        self.DGE_gene_umis = defaultdict(set)
        self.DGE_gene_cells = defaultdict(set)

        # flag to enable consistency checking (for debugging)
        self._assert = _assert
        self.channels = set([main_channel])

    def add_read(self, gene, cell, umi, channel="count"):
        self.DGE_cells.add(cell)
        self.DGE_genes.add(gene)

        # margin counters for validation
        if channel == self.main_channel:
            self.DGE_cell_reads[cell] += 1

        if (gene, umi) in self.DGE_cell_umis[cell]:
            dup = True
        else:
            dup = False

        self.DGE_cell_umis[cell].add((gene, umi))
        if self._assert:
            self.DGE_cell_genes[cell].add(gene)
            self.DGE_gene_reads[gene] += 1
            self.DGE_gene_umis[gene].add((cell, umi))
            self.DGE_gene_cells[gene].add(cell)

        # main count matrices
        self.DGE_umis[(gene, cell)][channel].add(umi)
        # if channel == self.main_channel:
        self.DGE_reads[(gene, cell)][channel] += 1

        self.channels.add(channel)

        return dup

    def make_DGEs(self):
        import scipy.sparse
        import anndata

        # count UMIs, reads across all channels into sparse arrays
        ind_of_cell = {}
        obs = []
        for cell in sorted(self.DGE_cells):
            ind_of_cell[cell] = len(obs)
            obs.append(cell)

        ind_of_gene = {}
        var = []
        for gene in sorted(self.DGE_genes):
            ind_of_gene[gene] = len(var)
            var.append(gene)

        umi_counts_by_channel = defaultdict(list)
        read_counts_by_channel = defaultdict(list)
        row_ind = []
        col_ind = []
        for (gene, cell), umi_dict in self.DGE_umis.items():
            for channel in self.channels:
                umis = umi_dict[channel]
                umi_counts_by_channel[channel].append(
                    len(umis)
                )  # <- the actual expression count
                read_counts_by_channel[channel].append(
                    self.DGE_reads[(gene, cell)][channel]
                )

            row_ind.append(ind_of_cell[cell])
            col_ind.append(ind_of_gene[gene])

        # print(
        #     ">>>DIMENSION CHECK",
        #     self.channels,
        #     len(self.DGE_umis.items()),
        #     [(k, len(v)) for k, v in umi_counts_by_channel.items()],
        #     len(umi_counts_by_channel[self.main_channel]),
        #     len(row_ind),
        #     len(col_ind),
        # )
        X = scipy.sparse.csr_matrix(
            (umi_counts_by_channel[self.main_channel], (row_ind, col_ind))
        )
        # convert to anndata object and store
        adata = anndata.AnnData(X)
        adata.obs_names = obs
        adata.var_names = var
        adata.layers[f"reads_{self.main_channel}"] = scipy.sparse.csr_matrix(
            (read_counts_by_channel[self.main_channel], (row_ind, col_ind))
        )

        for channel in self.channels:
            if channel == self.main_channel:
                continue

            umi_counts = umi_counts_by_channel[channel]

            adata.layers[channel] = scipy.sparse.csr_matrix(
                (umi_counts, (row_ind, col_ind))
            )
            read_counts = read_counts_by_channel[channel]
            adata.layers[f"reads_{channel}"] = scipy.sparse.csr_matrix(
                (read_counts, (row_ind, col_ind))
            )
            # Do we want to keep the extra UMI or read counts?
            # adata.layers[name] = aextra.layers["reads"]

        # print(adata)
        return adata

    def check_DGEs_vs_margin_counts(self, ann_umis, ann_reads):
        import numpy as np

        if not self._assert:
            raise ValueError(
                "DGE was not initialized with _assert = True, which is required for check_DGEs_vs_margin_counts()"
            )

        cell_names = ann_umis.obs_names
        gene_names = ann_umis.var_names

        def sets_to_vec(S, names):
            return np.array([len(S[n]) for n in names])

        def count_to_vec(S, names):
            return np.array([S[n] for n in names])

        def flatvec(m):
            return np.array(m).ravel()

        assert (
            flatvec(ann_umis.X.sum(axis=1))
            == sets_to_vec(self.DGE_cell_umis, cell_names)
        ).all()
        assert (
            flatvec(ann_umis.X.sum(axis=0))
            == sets_to_vec(self.DGE_gene_umis, gene_names)
        ).all()

        assert (
            flatvec(ann_reads.X.sum(axis=1))
            == count_to_vec(self.DGE_cell_reads, cell_names)
        ).all()
        assert (
            flatvec(ann_reads.X.sum(axis=0))
            == count_to_vec(self.DGE_gene_reads, gene_names)
        ).all()

        detect_reads = ann_reads.X > 0
        assert (
            flatvec(detect_reads.sum(axis=1))
            == sets_to_vec(self.DGE_cell_genes, cell_names)
        ).all()
        assert (
            flatvec(detect_reads.sum(axis=0))
            == sets_to_vec(self.DGE_gene_cells, gene_names)
        ).all()

        detect_umis = ann_umis.X > 0
        assert (
            flatvec(detect_umis.sum(axis=1))
            == sets_to_vec(self.DGE_cell_genes, cell_names)
        ).all()
        assert (
            flatvec(detect_umis.sum(axis=0))
            == sets_to_vec(self.DGE_gene_cells, gene_names)
        ).all()


# def assess_reads_per_UMI(ann_umis, ann_reads):
#     rpu = (
#         np.array(ann_reads.X.sum(axis=1), dtype=float).ravel()
#         / np.array(ann_umis.X.sum(axis=1), dtype=float).ravel()
#     )
#     perc = np.percentile(rpu, [5, 25, 50, 75, 95])
#     # print(perc)
#     return perc


def parse_cmdline():
    parser = argparse.ArgumentParser(
        description="quantify miRNA expression from BAM files"
    )

    parser.add_argument(
        "bam_in",
        help="bam input (default=stdin)",
        default="/dev/stdin",
        # nargs="+",
    )
    parser.add_argument(
        "--sample-name",
        help="sample identifier (default='NA')",
        default="NA",
    )

    parser.add_argument(
        "--parse-gn",
        help="parse gf,gn BAM tags instead of mapping information",
        default=False,
        action="store_true"
        # nargs="+",
    )
    parser.add_argument(
        "--ignore-gf",
        help="which gene functions to ignore? comma-separated string. default=INTRONIC,INTERGENIC",
        default="INTRONIC,INTERGENIC",
        # nargs="+",
    )

    parser.add_argument(
        "--translate",
        help="translate gene-names using this two-column lookup table w columns 'original' and 'target'",
        default="",
        # nargs="+",
    )
    parser.add_argument(
        "--skim",
        help="skim through the BAM by investigating only every <skim>-th record (default=1 off)",
        default=1,
        type=int
        # nargs="+",
    )
    parser.add_argument(
        "--output-counted",
        help="direct neg control alignments that are counted to this BAM file. default=off",
        default="",
    )
    parser.add_argument(
        "--output-discarded",
        help="direct reads that are discarded for reasons other than PCR-duplicate to this BAM file. default=off",
        default="",
    )
    parser.add_argument(
        "--output-passed",
        help="direct reads that pass all filters (but may be PCR-duplicates) to this BAM file. default=off",
        default="",
    )

    parser.add_argument(
        "--output-detailed",
        help="direct counts for individual miRNA family members to this file default=off",
        default="",
    )
    parser.add_argument(
        "--output-DGE",
        help="output a single-cell digital gene expression matrix with UMI COUNTS",
        default="",
    )
    parser.add_argument(
        "--layers",
        default="reads",
        help="comma-separated list of extra info to add as layers",
    )
    # parser.add_argument(
    #     "--output-DGE-reads",
    #     help="output a single-cell digital gene expression matrix with READ COUNTS",
    #     default="",
    # )
    parser.add_argument(
        "--count-func",
        default="ligation_product",
        help="the function which decides whether to count a read",
        choices=["ligation_product", "targeted_primer", "everything"],
    )
    parser.add_argument(
        "--min-match",
        default=17,
        type=int,
        help="minimum number of matching bases required (default=17)",
    )
    parser.add_argument(
        "--max-clipped",
        default=50,
        type=int,
        help="maximum number of soft-clipped 3'bases allowed (default=50)",
    )
    parser.add_argument(
        "--min-BC",
        default=0,
        type=int,
        help="expect at least this number of barcode bases at the 5' end (default=0)",
    )
    parser.add_argument(
        "--name-filter",
        default="",
        help="require presence of this string in qname (default='' -> off)",
    )
    parser.add_argument(
        "--name-exclude",
        default="",
        help="discard if this string is present in qname (default='' -> off)",
    )

    return parser.parse_args()


def count_ligation_product(ca):
    return (
        not (ca.short or ca.primer) and "A3_has_polyA" in ca.flags and ca.gene != "NA"
    )


def count_targeted_primer(ca):
    return ca.primer


def count_everything(ca):
    return True


if __name__ == "__main__":
    args = parse_cmdline()

    lkup = {}
    if args.translate:
        for row in pd.read_csv(args.translate, sep="\t").itertuples():
            lkup[row.original] = row.target

    if args.parse_gn:
        gene_mode = "gn_tag"
    else:
        gene_mode = "chrom"

    ignore = args.ignore_gf.split(",")

    count_func = {
        "ligation_product": count_ligation_product,
        "targeted_primer": count_targeted_primer,
        "everything": count_everything,
    }[args.count_func]
    dge = DGE()
    for ca in AlignmentClassifier(
        args.sample_name,
        args.bam_in,
        chrom_to_gene=lkup,
        gene_assign_mode=gene_mode,
        ignore_gf=ignore,
    ):
        ca = (
            # .check_qname(keywords=["TSO", "polyA"])
            ca.check_tags(
                A3=["polyA"], A5=["TSO"]
            ).check_CIGAR()  # fills flags["A3_has_polyA"] etc.  # fills .clip5, .n_match, .clip3
        )

        ca.short = ca.n_match < args.min_match
        ca.reverse = ca.aln.is_reverse
        # ca.PCR_dup  = ca.
        ca.primer = ("A5_has_TSO" in ca.flags) or ("A3_missing_polyA" in ca.flags)
        ca.count = count_func(ca)
        # print(ca.n_match, "flags=", ca.flags)
        # print(ca.short, ca.primer, ca.reverse, ca.count)

        channels = ["count", "short", "reverse", "primer"]
        for c in channels:
            if getattr(ca, c):
                # print(ca.aln)
                # print("counting as", c)
                dge.add_read(gene=ca.gene, cell=ca.cell, umi=ca.umi, channel=c)

        # print("next")
    if args.output_DGE and len(dge.DGE_reads):
        adata = dge.make_DGEs()
        # dge.check_DGEs_vs_margin_counts(ann_umis, ann_reads)
        # print("storing AnnData object")
        adata.write(args.output_DGE)
        # ann_reads.write(args.output_DGE_reads)

        # sys.stderr.write(
        #     f"### reads-to-UMI ratio quartiles: {assess_reads_per_UMI(ann_umis, ann_reads)} \n"
        # )

    # # print("starting")
    # counts = defaultdict(int)
    # discarded = defaultdict(lambda: defaultdict(int))  # key is [reason][miRNA] -> count
    # counts_detailed = defaultdict(int)
    # discarded_detailed = defaultdict(lambda: defaultdict(int))
    # rev = defaultdict(int)
    # rev_detailed = defaultdict(int)
    # stats = defaultdict(lambda: defaultdict(int))
    # stats_detailed = defaultdict(lambda: defaultdict(int))
    # clipped = defaultdict(lambda: defaultdict(int))
    # nclipped = defaultdict(int)
    # clipped5 = defaultdict(lambda: defaultdict(int))
    # nclipped5 = defaultdict(int)

    # sam = pysam.AlignmentFile(args.bam_in, "rb", check_sq=False)
    # if args.output_counted:
    #     sam_counted = pysam.AlignmentFile(args.output_counted, "wb", template=sam)
    # else:
    #     sam_counted = None

    # if args.output_discarded:
    #     sam_discard = pysam.AlignmentFile(args.output_discarded, "wb", template=sam)
    # else:
    #     sam_discard = None

    # if args.output_passed:
    #     sam_passed = pysam.AlignmentFile(args.output_passed, "wb", template=sam)
    # else:
    #     sam_passed = None

    # full_names = set()
    # simple_names = set()
    # dge = DGE()
    # layer_names = set(args.layers.split(","))
    # # print(f"DGE-counting the following events: {layer_names}")

    # def record_dge_extra(full, reason, read):
    #     if reason in layer_names:
    #         dge.add_read(full, read.get_tag("CB"), read.get_tag("MI"), channel=reason)

    # def flag_drop(simple, full, reason, read):
    #     stats[simple][f"n_{reason}"] += 1
    #     stats_detailed[full][f"n_{reason}"] += 1
    #     discarded[reason][simple] += 1
    #     discarded_detailed[reason][full] += 1

    #     record_dge_extra(full, reason, read)
    #     record_dge_extra(full, "dropped", read)

    #     if reason != "PCR_duplicate" and sam_discard:
    #         key = read.query_sequence + read.get_tag("CB") + read.get_tag("MI")
    #         # if not key in discarded_uniq:
    #         read.set_tag("XD", reason)
    #         sam_discard.write(read)

    #         # discarded_uniq.add(key)

    #     return True

    # for i_read, read in enumerate(sam.fetch(until_eof=True)):
    #     if (i_read % args.skim) != 0:
    #         continue

    #     # print(read)
    #     full, simple, L = get_mirname(sam, read, args.parse_gf)
    #     full_names.add(full)
    #     simple_names.add(simple)
    #     stats[simple]["n_reads"] += 1
    #     stats_detailed[full]["n_reads"] += 1
    #     if full != simple:
    #         stats[full]["n_reads"] += 1
    #     drop = False

    #     n_match = 0
    #     if args.min_match:
    #         for op, n in read.cigartuples:
    #             # find the largest contiguous stretch of aligned bases
    #             if op == 0:
    #                 n_match = max(n_match, n)

    #     if n_match < args.min_match:
    #         drop = flag_drop(simple, full, "match_too_short", read)

    #     n5 = 0
    #     if read.cigartuples:
    #         op, n = read.cigartuples[0]
    #         if op == 4:
    #             n5 = n

    #     if "TSO" in read.qname:
    #         if full != simple:
    #             clipped5[full]["TATGGG"] += 1
    #         clipped5[simple]["TATGGG"] += 1
    #         if full != simple:
    #             nclipped5[full] += 1
    #         nclipped5[simple] += 1
    #     elif n5:
    #         n_clipped = n5
    #         if full != simple:
    #             clipped5[full][read.query_sequence[:n]] += 1
    #         clipped5[simple][read.query_sequence[:n]] += 1
    #         if full != simple:
    #             nclipped5[full] += 1
    #         nclipped5[simple] += 1

    #     n_clipped = 0
    #     if read.cigartuples:
    #         op, n = read.cigartuples[-1]
    #         if op == 4:
    #             clipped_seq = clip_pA(read.query_sequence[-n:])
    #             n_clipped = len(clipped_seq)
    #             if full != simple:
    #                 clipped[full][clipped_seq] += 1
    #             clipped[simple][clipped_seq] += 1
    #         else:
    #             if full != simple:
    #                 clipped[full][""] += 1
    #             clipped[simple][""] += 1

    #     if full != simple:
    #         nclipped[full] += 1
    #     nclipped[simple] += 1
    #     clipped[simple]["all_occurrences"] += 1
    #     if full != simple:
    #         clipped[full]["all_occurrences"] += 1

    #     if n_clipped > args.max_clipped:
    #         drop = flag_drop(simple, full, "too_much_clipped", read)

    #     if args.min_BC and (n5 != args.min_BC):
    #         drop = flag_drop(simple, full, "missing_BC", read)

    #     if args.name_filter:
    #         if not args.name_filter in read.qname:
    #             drop = flag_drop(simple, full, f"name_missing_{args.name_filter}", read)
    #             # stats[simple]["n_name_filtered"] += 1

    #             # stats_detailed[full]["n_name_filtered"] += 1
    #             # discarded[f"name_missing_{args.name_filter}"][simple] += 1
    #             # discarded_detailed[f"name_missing_{args.name_filter}"][full] += 1

    #             # drop = True

    #     if args.name_exclude:
    #         if args.name_exclude in read.qname:
    #             drop = flag_drop(simple, full, f"name_has_{args.name_exclude}", read)

    #             # stats[simple]["n_name_excluded"] += 1
    #             # stats_detailed[full]["n_name_excluded"] += 1
    #             # discarded[f"name_has_{args.name_exclude}"][simple] += 1
    #             # discarded_detailed[f"name_has_{args.name_exclude}"][full] += 1

    #             # drop = True

    #     if read.is_reverse:
    #         drop = flag_drop(simple, full, "reverse", read)

    #         # stats[simple]["n_reverse"] += 1
    #         # stats_detailed[full]["n_reverse"] += 1
    #         # discarded["reverse"][simple] += 1
    #         # discarded_detailed["reverse"][full] += 1

    #     if not drop:
    #         # dup = dge.add_read(full, read.get_tag("CB"), read.get_tag("MI"))
    #         # if dup:
    #         #     drop = flag_drop(simple, full, "PCR_duplicate")
    #         if sam_passed:
    #             sam_passed.write(read)

    #         if args.output_DGE:
    #             dup = dge.add_read(full, read.get_tag("CB"), read.get_tag("MI"))
    #             if dup:
    #                 drop = flag_drop(simple, full, "PCR_duplicate", read)

    #     if drop:
    #         stats[simple]["n_dropped"] += 1
    #         stats_detailed[full]["n_dropped"] += 1
    #     else:
    #         counts[simple] += 1
    #         counts_detailed[full] += 1
    #         if simple.startswith("NegCtrl") and sam_counted:
    #             sam_counted.write(read)

    #         stats[simple]["n_kept"] += 1
    #         stats_detailed[full]["n_kept"] += 1

    # if args.output_DGE and len(dge.DGE_reads):
    #     adata = dge.make_DGEs()
    #     # dge.check_DGEs_vs_margin_counts(ann_umis, ann_reads)
    #     # print("storing AnnData object")
    #     adata.write(args.output_DGE)
    #     # ann_reads.write(args.output_DGE_reads)

    #     # sys.stderr.write(
    #     #     f"### reads-to-UMI ratio quartiles: {assess_reads_per_UMI(ann_umis, ann_reads)} \n"
    #     # )

    # out_counts_bulk(sys.stdout, counts, discarded, stats)
    # # for mir, count in sorted(counts.items(), key=lambda x: -x[1]):
    # #     # print(f"{mir}\t{sam.get_reference_length(mir)}\t{count}\t{rev[mir]}")
    # #     print(
    # #         f"{mir}\t-1\t{count}\t{rev[mir]}\t{stats[mir]['n_dropped']}\t{stats[mir]['n_dropped']/count:.1f}"
    # #     )

    # if args.output_detailed:
    #     out_counts_bulk(
    #         open(args.output_detailed, "wt"),
    #         counts_detailed,
    #         discarded_detailed,
    #         stats_detailed,
    #     )

    # for simple_name in simple_names:
    #     data = stats[simple_name]
    #     for key, value in sorted(data.items()):
    #         perc = 100 * value / data["n_reads"]
    #         sys.stderr.write(f"{simple_name}\t{key}\t{value}\t{perc:.2f} %\n")

    # # print the most common clipped sequences
    # sys.stderr.write("### most common 3' non-miRNA sequences detected\n")
    # for mirname, cdata in sorted(clipped.items()):
    #     # cdata = clipped[simple_name]
    #     sorted_counts = sorted(cdata.items(), key=lambda x: -x[1])
    #     for seq, n in sorted_counts[:50]:
    #         # if n > 9:
    #         sys.stderr.write(
    #             f"{'simple' if mirname in simple_names else 'full'}\t3'\t{mirname}\t{n}\t{100 * n/nclipped[mirname]:.2f} %\t{seq}\n"
    #         )
    #     if len(sorted_counts) >= 50:
    #         n_other = 0
    #         for seq, n in sorted_counts[50:]:
    #             n_other += n

    #         sys.stderr.write(
    #             f"{'simple' if mirname in simple_names else 'full'}\t3'\t{mirname}\t{n_other}\t{100 * n_other/nclipped[mirname]:.2f} %\tother\n"
    #         )

    # # print the most common clipped sequences
    # sys.stderr.write("### most common 5' non-miRNA sequences detected\n")
    # for mirname, cdata in sorted(clipped5.items()):
    #     # cdata = clipped[simple_name]
    #     for seq, n in sorted(cdata.items(), key=lambda x: -x[1])[:50]:
    #         if n > 9:
    #             sys.stderr.write(
    #                 f"{'simple' if mirname in simple_names else 'full'}\t5'\t{mirname}\t{n}\t{100 * n/nclipped5[mirname]:.2f} %\t{seq}\n"
    #             )



import logging

logger_name = "spacemake.main"
logger = logging.getLogger(logger_name)


class Spacemake:
    """Spacemake.

    Class to access spacemake processed data from python.

    """

    def __init__(self, root):
        """__init__ constructor function of the Spacemake class

        :param root: Path to the spacemake root directory.
        :type root: str
        """
        from spacemake.config import get_global_config
        from spacemake.project_df import get_global_ProjectDF

        self.root = root
        self.config = get_global_config(root)
        self.project_df = get_global_ProjectDF(root)

    def load_processed_adata(
        self, project_id, sample_id, run_mode_name, umi_cutoff
    ): #-> anndata.AnnData:
        """Load spacemake processed data.

        :param project_id: project_id of the data to be loaded.
        :type project_id: str
        :param sample_id: sample_id of the data to be loaded.
        :type sample_id: str
        :param run_mode_name: name of the run mode of the data to be loaded.
            Each sample can have several run_modes during sample addition,
            here only one option needs to be provided.
        :type run_mode_name: str
        :param umi_cutoff: the umi_cutoff of the data to be loaded. Each
            run_mode can have several umi_cutoffs provided during configuration
            here only one option needs to be provided.
        :type umi_cutoff: int
        :returns: A spacemake processed and analyzed AnnData object, containing
            the results of the analysis.
        :rtype: anndata.AnnData
        """
        import scanpy as sc
        # import anndata


        self.project_df.assert_run_mode(project_id, sample_id, run_mode_name)
        run_mode = self.config.get_run_mode(run_mode_name)

        if not int(umi_cutoff) in [int(uc) for uc in run_mode.variables["umi_cutoff"]]:
            raise SpacemakeError(
                f"run_mode={run_mode} has no " + f"umi_cutoff={umi_cutoff}"
            )

        adata_raw = self.load_raw_spatial_adata(
            project_id=project_id, sample_id=sample_id, run_mode_name=run_mode_name
        )

        adata = sc.read(
            f"{self.root}/projects/{project_id}/processed_data/{sample_id}/"
            + f"illumina/complete_data/automated_analysis/{run_mode_name}/"
            + f"umi_cutoff_{umi_cutoff}/results.h5ad"
        )

        if "run_mode_variables" not in adata.uns.keys():
            adata.uns["run_mode_variables"] = run_mode.variables
        if "puck_variables" not in adata.uns.keys():
            adata.uns["puck_variables"] = adata_raw.uns["puck_variables"]

        return adata

    def load_raw_spatial_adata(
        self, project_id, sample_id, run_mode_name
    ): #-> anndata.AnnData:
        """Load raw, spacemake processed data.

        This function will load the raw countr matrix, created by spacemake.

        :param project_id: project_id of the raw data to be loaded.
        :type project_id: str
        :param sample_id: sample_id of the raw data to be loaded.
        :type sample_id: str
        :param run_mode_name: name of the run mode of the raw data to be loaded.
            Each sample can have several run_modes during sample addition,
            here only one option needs to be provided.
        :type run_mode_name: str
        :returns: A spacemake processed AnnData object, containing unfiltered
            raw expression data, and all cells or spatial units in the dataset.
        :rtype: anndata.AnnData
        """
        import scanpy as sc

        self.project_df.assert_run_mode(project_id, sample_id, run_mode_name)
        run_mode = self.config.get_run_mode(run_mode_name)

        dge_type = ""
        dge_cleaned = ""
        polyA_adapter_trimmed = ""
        mm_included = ""

        if run_mode.variables["polyA_adapter_trimming"]:
            polyA_adapter_trimmed = ".polyA_adapter_trimmed"

        if run_mode.variables["count_intronic_reads"]:
            dge_type = ".all"
        else:
            dge_type = ".exon"

        if run_mode.variables["count_mm_reads"]:
            mm_included = ".mm_included"

        if run_mode.variables["clean_dge"]:
            dge_cleaned = ".cleaned"

        adata = sc.read(
            f"{self.root}/projects/{project_id}/processed_data/{sample_id}/"
            + f"illumina/complete_data/dge/dge{dge_type}{dge_cleaned}"
            + f"{polyA_adapter_trimmed}{mm_included}.spatial_beads.h5ad"
        )

        if "puck_variables" not in adata.uns.keys():
            from spacemake.preprocess import attach_puck_variables

            adata = attach_puck_variables(
                adata,
                puck_variables=self.project_df.get_puck_variables(
                    project_id=project_id, sample_id=sample_id
                ),
            )

        if "run_mode_variables" not in adata.uns.keys():
            adata.uns["run_mode_variables"] = run_mode.variables

        return adata


def get_novosparc_variables(pdf, args):
    """get_novosparc_variables.

    :param pdf:
    :param args:
    """
    # assert that sample exists
    pdf.assert_sample(args["project_id"], args["sample_id"])

    def populate_variables_from_args(pdf, args, arg_prefix=""):
        """populate_variables_from_args.

        :param pdf:
        :param args:
        :param arg_prefix:
        """
        # get sample info
        sample_info = pdf.get_sample_info(
            project_id=args[f"{arg_prefix}project_id"],
            sample_id=args[f"{arg_prefix}sample_id"],
        )

        # populate return dictionary
        ret = {
            f"{arg_prefix}project_id": args[f"{arg_prefix}project_id"],
            f"{arg_prefix}sample_id": args[f"{arg_prefix}sample_id"],
        }

        # get run mode
        if f"{arg_prefix}run_mode" in args:
            ret[f"{arg_prefix}run_mode"] = args[f"{arg_prefix}run_mode"]
        else:
            run_mode_name = sample_info["run_mode"][0]
            ret[f"{arg_prefix}run_mode"] = run_mode_name
            logger.info(f"No run_mode provided, using {run_mode_name}")

        run_mode = pdf.config.get_run_mode(ret[f"{arg_prefix}run_mode"])

        if f"{arg_prefix}umi_cutoff" not in args:
            umi_cutoff = run_mode.variables["umi_cutoff"][0]
            ret[f"{arg_prefix}umi_cutoff"] = umi_cutoff
            logger.info(f"No umi_cutoff provided, using {umi_cutoff}")
        else:
            ret[f"{arg_prefix}umi_cutoff"] = args[f"{arg_prefix}umi_cutoff"]

        return ret

    ret = populate_variables_from_args(pdf, args)

    if "reference_project_id" not in args or "reference_sample_id" not in args:
        logger.info(
            "No reference_project_id or reference_sample_id provided,"
            + " running novosparc de-novo..."
        )
        ret["reference_project_id"] = ""
        ret["reference_sample_id"] = ""
        ret["reference_umi_cutoff"] = ""
        ret["reference_run_mode"] = ""
    else:
        pdf.assert_sample(args["reference_project_id"], args["reference_sample_id"])

        logger.info(
            "Using (project_id, sample_id)="
            + f"({args['reference_project_id']}, {args['reference_sample_id']})"
            + " reference, running novosparc with reference..."
        )

        novosparc_ret = populate_variables_from_args(pdf, args, arg_prefix="reference_")

        ret = {**ret, **novosparc_ret}

    return ret


_spacemake_instance = None


def get_spacemake_object():
    global _spacemake_instance
    if _spacemake_instance is None:
        _spacemake_instance = Spacemake(".")

    return _spacemake_instance


# def get_ConfigFile():
#     spmk = get_spacemake_object()
#     return spmk.config


# def get_ProjectDF():
#     spmk = get_spacemake_object()
#     return spmk.project_df



# from anndata import AnnData
from spacemake.preprocess.dge import calculate_adata_metrics


def compute_neighbors(adata, min_dist=None, max_dist=None):
    """Compute all direct neighbors of all spots in the adata. Currently
        tailored for 10X Visium.
    Args:
        adata: an AnnData object
        min_dist: int, minimum distance to consider neighbors
        max_dist: int, maximum distance to consider neighbors
    Returns:
        neighbors: a dictionary holding the spot IDs for every spot
    """
    import numpy as np
    from scipy.spatial.distance import cdist

    # Calculate all Euclidean distances on the adata
    dist_mtrx = cdist(adata.obsm["spatial"], adata.obsm["spatial"])
    neighbors = dict()

    for i in range(adata.obs.shape[0]):
        neighbors[i] = np.where(
            (dist_mtrx[i, :] > min_dist) & (dist_mtrx[i, :] < max_dist)
        )[0]

    return neighbors


def compute_islands(adata, min_umi):
    """Find contiguity islands

    Args:
        adata: an AnnData object
        cluster: a cell type label to compute the islands for
    Returns:
        islands: A list of lists of spots forming contiguity islands
    """
    import numpy as np
    import itertools

    # this is hard coded for now for visium, to have 6 neighbors per spot
    # TODO: define an iterative approach where the key is to have around 6
    # neighbors per spot on average
    neighbors = compute_neighbors(adata, min_dist=0, max_dist=3)
    spots_cluster = np.where(np.array(adata.obs["total_counts"]) < min_umi)[0]

    # create a new dict holding only neighbors of the cluster
    islands = []

    for spot in neighbors:
        if spot not in spots_cluster:
            continue
        islands.append({spot}.union({x for x in neighbors[spot] if x in spots_cluster}))

    # merge islands with common spots
    island_spots = set(itertools.chain.from_iterable(islands))

    for each in island_spots:
        components = [x for x in islands if each in x]
        for i in components:
            islands.remove(i)
        islands += [list(set(itertools.chain.from_iterable(components)))]

    return islands


def nonsingular(vmin, vmax, expander=0.001, tiny=1e-15, increasing=True):
    """
    Modify the endpoints of a range as needed to avoid singularities.

    This code was adapted from matplotlib.transforms
    "Copyright (c) 2012- Matplotlib Development Team; All Rights Reserved"

    Args
        vmin, vmax: float, the initial endpoints.
        expander: float, default: 0.001, fractional amount by which *vmin* and
                  *vmax* are expanded if the original interval is too small,
                  based on *tiny*.
        tiny : float, default: 1e-15, hreshold for the ratio of the interval to
            the maximum absolute value of its endpoints.
            If the interval is smaller than this, it will be expanded.
            This value should be around 1e-15 or larger;
            otherwise the interval will be approaching the double precision
            resolution limit.
        increasing: bool, default: True, if True, swap *vmin*, *vmax* if *vmin* > *vmax*.
    Returns:
        vmin, vmax: float, endpoints, expanded and/or swapped if necessary.
                    If either input is inf or NaN, or if both inputs are 0 or very
                    close to zero, it returns -*expander*, *expander*.
    """
    import numpy as np

    if (not np.isfinite(vmin)) or (not np.isfinite(vmax)):
        return -expander, expander

    swapped = False
    if vmax < vmin:
        vmin, vmax = vmax, vmin
        swapped = True

    vmin, vmax = map(float, [vmin, vmax])

    maxabsvalue = max(abs(vmin), abs(vmax))
    if maxabsvalue < (1e6 / tiny) * np.finfo(float).tiny:
        vmin = -expander
        vmax = expander

    elif vmax - vmin <= maxabsvalue * tiny:
        if vmax == 0 and vmin == 0:
            vmin = -expander
            vmax = expander
        else:
            vmin -= expander * abs(vmin)
            vmax += expander * abs(vmax)

    if swapped and not increasing:
        vmin, vmax = vmax, vmin
    return vmin, vmax


def detect_tissue(adata, min_umi):
    """Detect tissue: first find beads with at least min_umi UMIs, then detect island in the rest

    Args
        adata: an AnnData object, with spatial coordinates
        min_umi: integer, the min umi to be assigned as tissue bead by default
    Returns:
        tissue_indices: a list of indices which should be kept for this AnnData object
    """
    import numpy as np

    islands = compute_islands(adata, min_umi)

    # find the sizes of the islands. remove the biggest, as either the tissue has a big hole in it
    # or there are not so many big islands in which case removal is OK.
    # to be evaluated later...
    island_sizes = [len(island) for island in islands]

    tissue_islands = np.delete(islands, np.argmax(island_sizes))

    # get the indices of the islands
    tissue_indices = np.where(np.array(adata.obs["total_counts"]) >= min_umi)[0]

    tissue_indices = np.append(tissue_indices, np.hstack(tissue_islands))

    adata = adata[tissue_indices, :]

    return adata


def create_mesh(width, height, diameter, distance, push_x=0, push_y=0):
    import numpy as np

    distance_y = np.sqrt(3) * distance

    x_coord = np.arange(push_x, width + diameter, distance)
    y_coord = np.arange(push_y, height + diameter, distance_y)

    X, Y = np.meshgrid(x_coord, y_coord)
    xy = np.vstack((X.flatten(), Y.flatten())).T

    return xy


def binning_hexagon(x, y, gridsize, extent=None, last_row=False):
    """Bins x,y points into a mesh of gridsize (across x axis, or xy axes).
    Points are assigned to the closest hexagon, without explicitly calculating
    pairwise distances.

    Does not require to create a mesh beforehand, this function handles the
    mesh and nearest neighbor.

    This code was adapted from matplotlib
    "Copyright (c) 2012- Matplotlib Development Team; All Rights Reserved"

    Args
        x, y: numpy.ndarray, x, y spatial coordinates of points
        gridsize: float or tuple, the amount of hexagons in x direction (float);
                  y direction is automatically computed; or specified if gridsize is
                  a tuple.
        extent: tuple, of (x_min, x_max, y_min, y_max)
        last_row: bool, whether a last row is created in mesh or not
    Returns:
        coordinates: numpy.ndarray, a (x_mesh x y_mesh) x 2 matrix, with the coordinates
                     (centres) of each hexagon in the binned mesh
        accumulated: list, contains the indices from the x, y arrays that were binned
                     to each hexagon in the mesh.
    """
    import numpy as np

    if np.iterable(gridsize):
        nx, ny = gridsize
    else:
        nx = gridsize
        ny = int(nx / np.sqrt(3))
    # Count the number of data in each hexagon
    x = np.asarray(x, float)
    y = np.asarray(y, float)

    # Will be log()'d if necessary, and then rescaled.
    tx = x
    ty = y

    if extent is not None:
        xmin, xmax, ymin, ymax = extent
    else:
        xmin, xmax = (tx.min(), tx.max()) if len(x) else (0, 1)
        ymin, ymax = (ty.min(), ty.max()) if len(y) else (0, 1)

        # to avoid issues with singular data, expand the min/max pairs
        xmin, xmax = nonsingular(xmin, xmax, expander=0.1)
        ymin, ymax = nonsingular(ymin, ymax, expander=0.1)

    nx1 = nx + 1
    if last_row:
        ny1 = ny + 1
    else:
        ny1 = ny
    nx2 = nx
    ny2 = ny
    n = nx1 * ny1 + nx2 * ny2

    # In the x-direction, the hexagons exactly cover the region from
    # xmin to xmax. Need some padding to avoid roundoff errors.
    padding = 1.0e-9 * (xmax - xmin)
    xmin -= padding
    xmax += padding
    sx = (xmax - xmin) / nx
    sy = (ymax - ymin) / ny

    # Positions in hexagon index coordinates.
    ix = (tx - xmin) / sx
    iy = (ty - ymin) / sy
    ix1 = np.round(ix).astype(int)
    iy1 = np.round(iy).astype(int)
    ix2 = np.floor(ix).astype(int)
    iy2 = np.floor(iy).astype(int)

    # flat indices, plus one so that out-of-range points go to position 0.
    i1 = np.where(
        (0 <= ix1) & (ix1 < nx1) & (0 <= iy1) & (iy1 < ny1), ix1 * ny1 + iy1 + 1, 0
    )
    i2 = np.where(
        (0 <= ix2) & (ix2 < nx2) & (0 <= iy2) & (iy2 < ny2), ix2 * ny2 + iy2 + 1, 0
    )

    d1 = (ix - ix1) ** 2 + 3.0 * (iy - iy1) ** 2
    d2 = (ix - ix2 - 0.5) ** 2 + 3.0 * (iy - iy2 - 0.5) ** 2
    bdist = d1 < d2

    Cs_at_i1 = [[] for _ in range(1 + nx1 * ny1)]
    Cs_at_i2 = [[] for _ in range(1 + nx2 * ny2)]
    for i in range(len(x)):
        if bdist[i]:
            Cs_at_i1[i1[i]].append(i)
        else:
            Cs_at_i2[i2[i]].append(i)

    accumulated = [acc for Cs_at_i in [Cs_at_i1, Cs_at_i2] for acc in Cs_at_i[1:]]

    coordinates = np.zeros((n, 2), float)
    coordinates[: nx1 * ny1, 0] = np.repeat(np.arange(nx1), ny1)
    coordinates[: nx1 * ny1, 1] = np.tile(np.arange(ny1), nx1)
    coordinates[nx1 * ny1 :, 0] = np.repeat(np.arange(nx2) + 0.5, ny2)
    coordinates[nx1 * ny1 :, 1] = np.tile(np.arange(ny2), nx2) + 0.5
    coordinates[:, 0] *= sx
    coordinates[:, 1] *= sy
    coordinates[:, 0] += xmin
    coordinates[:, 1] += ymin

    return coordinates, accumulated


def aggregate_adata_by_indices(
    adata, idx_to_aggregate, idx_aggregated, coordinates_aggregated
):
    import pandas as pd
    import numpy as np
    import anndata

    from scipy.sparse import csr_matrix, csc_matrix, vstack, dok_matrix

    joined_C = adata.X[idx_to_aggregate]

    # at which indices does the index in the newly created matrix change
    change_ix = np.where(idx_aggregated[:-1] != idx_aggregated[1:])[0] + 1

    # array of indices, split by which row they should go together

    ix_array = np.asarray(
        np.split(np.arange(idx_aggregated.shape[0]), change_ix, axis=0), dtype="object"
    )

    joined_C_sumed = vstack(
        [
            csr_matrix(joined_C[ix_array[n].astype(int), :].sum(0))
            for n in range(len(ix_array))
        ]
    )

    aggregated_adata = anndata.AnnData(
        csc_matrix(joined_C_sumed),
        obs=pd.DataFrame(
            {
                "x_pos": coordinates_aggregated[:, 0],
                "y_pos": coordinates_aggregated[:, 1],
            }
        ),
        var=adata.var,
    )

    aggregated_adata.obsm["spatial"] = coordinates_aggregated

    # rename index
    aggregated_adata.obs.index.name = "cell_bc"

    def summarise_adata_obs_column(adata, column, summary_fun=sum):
        vals_to_join = adata.obs[column].to_numpy()[idx_to_aggregate]
        vals_joined = np.array(
            [
                summary_fun(vals_to_join[ix_array[n].astype(int)])
                for n in range(len(ix_array))
            ]
        )
        return vals_joined

    print(adata)

    # summarise and attach n_reads, calculate metrics (incl. pcr)
    calculate_adata_metrics(
        aggregated_adata,
        # provide the n_reads as a parameter
        n_reads=summarise_adata_obs_column(adata, "n_reads"),
    )

    aggregated_adata.obs["n_joined"] = [len(x) for x in ix_array]
    print(f"ix_array={ix_array} shape={ix_array.shape} dtype={ix_array.dtype}")
    print(f"idx_to_aggregate={idx_to_aggregate}")
    joined_dict = {i: idx_to_aggregate[x.astype(int)] for i, x in enumerate(ix_array)}

    indices_joined_spatial_units = dok_matrix(
        (len(joined_dict), len(adata.obs_names)), dtype=np.uint16
    )

    for obs_name_aggregate, obs_name_to_aggregate in joined_dict.items():
        indices_joined_spatial_units[obs_name_aggregate, obs_name_to_aggregate] = 1

    indices_joined_spatial_units = indices_joined_spatial_units.tocsr()
    aggregated_adata.uns["spatial_units_obs_names"] = np.array(adata.obs_names)
    aggregated_adata.uns["indices_joined_spatial_units"] = indices_joined_spatial_units
    
    # carry the variables for the original puck variables
    if "puck_variables" in adata.uns:
        aggregated_adata.uns["puck_variables"] = adata.uns["puck_variables"]

    from statistics import mean

    for column in [
        "exact_entropy",
        "theoretical_entropy",
        "exact_compression",
        "theoretical_compression",
    ]:
        aggregated_adata.obs[column] = summarise_adata_obs_column(adata, column, mean)

    aggregated_adata.obs['exact_entropy'] = aggregated_adata.obs['exact_entropy'].astype(np.float16)
    aggregated_adata.obs['theoretical_entropy'] = aggregated_adata.obs['theoretical_entropy'].astype(np.float16)
    aggregated_adata.obs['exact_compression'] = aggregated_adata.obs['exact_compression'].astype(np.uint16)
    aggregated_adata.obs['theoretical_compression'] = aggregated_adata.obs['theoretical_compression'].astype(np.uint16)

    aggregated_adata.obs['n_counts'] = summarise_adata_obs_column(adata, "n_counts")
    return aggregated_adata


def create_meshed_adata(
    adata,
    px_by_um,
    spot_diameter_um=55,
    spot_distance_um=100,
    bead_diameter_um=10,
    mesh_type="circle",
    start_at_minimum=False,
    optimized_binning=True,
):
    import numpy as np

    from sklearn.metrics.pairwise import euclidean_distances

    if not mesh_type in ["circle", "hexagon"]:
        raise ValueError(f"unrecognised mesh type {mesh_type}")

    if mesh_type == "hexagon":
        spot_diameter_um = np.sqrt(3) * spot_diameter_um
        spot_distance_um = spot_diameter_um

    coords = adata.obsm["spatial"]

    if start_at_minimum:
        top_left_corner = np.min(coords, axis=0)
    else:
        top_left_corner = [0, 0]

    bottom_right_corner = np.max(coords, axis=0)
    width_px = abs(top_left_corner[0] - bottom_right_corner[0])
    height_px = abs(top_left_corner[1] - bottom_right_corner[1])

    # capture area width
    spot_radius_um = spot_diameter_um / 2

    um_by_px = 1.0 / px_by_um
    width_um = um_by_px * width_px
    spot_diameter_px = spot_diameter_um / um_by_px
    spot_radius_px = spot_radius_um / um_by_px
    spot_distance_px = spot_distance_um / um_by_px

    height_um = height_px * um_by_px

    # create meshgrid with one radius push
    xy = create_mesh(
        width_um,
        height_um,
        spot_diameter_um,
        spot_distance_um,
        push_x=spot_radius_um,
        push_y=spot_radius_um,
    )
    # create pushed meshgrid, pushed by
    xy_pushed = create_mesh(
        width_um,
        height_um,
        spot_diameter_um,
        spot_distance_um,
        push_x=spot_radius_um + spot_distance_um / 2,
        push_y=spot_radius_um + np.sqrt(3) * spot_distance_um / 2,
    )

    mesh = np.vstack((xy, xy_pushed))
    mesh_px = mesh / um_by_px
    # add the top_left_corner
    mesh_px = mesh_px + top_left_corner

    # example: the diameter of one visium spot is 55um. if we take any bead, which center
    # falls into that, assuming that a bead is 10um, we would in fact take all beads
    # within 65um diameter. for this reason, the max distance should be 45um/2 between
    # visium spot center and other beads
    max_distance_px = (spot_diameter_um - bead_diameter_um) / um_by_px / 2

    def _create_optimized_hex_mesh_properties(mesh_px):
        _y_values = np.unique(mesh_px[:, 1])
        _x_values = np.unique(mesh_px[:, 0])
        grid_x = len(mesh_px[mesh_px[:, 1] == _y_values[1]])
        grid_y = len(mesh_px[mesh_px[:, 0] == _x_values[1]])

        # Calculating the extent
        if len(mesh_px[mesh_px[:, 0] == _x_values[0]]) == grid_y:
            _offset = np.diff(_y_values[0:2])
            _last_row = False
        else:
            _offset = 0
            _last_row = True

        _extent = (
            mesh_px[:, 0].min(),
            mesh_px[:, 0].max(),
            mesh_px[:, 1].min(),
            mesh_px[:, 1].max() + _offset,
        )

        return grid_x, grid_y, _extent, _last_row

    if mesh_type == "circle":
        if optimized_binning:
            grid_x, grid_y, _extent, _last_row = _create_optimized_hex_mesh_properties(
                mesh_px
            )
            mesh_px, accum = binning_hexagon(
                coords[:, 0],
                coords[:, 1],
                gridsize=(grid_x, grid_y),
                extent=_extent,
                last_row=_last_row,
            )

            new_ilocs = [[i] * len(accum[i]) for i in range(len(accum))]
            new_ilocs = np.array([n for new_iloc in new_ilocs for n in new_iloc])
            original_ilocs = np.concatenate(accum).astype(int)

            distance_filter = (
                np.linalg.norm(
                    np.array(coords[original_ilocs]) - np.array(mesh_px[new_ilocs]),
                    axis=1,
                )
                < max_distance_px
            )

            new_ilocs = new_ilocs[distance_filter]
            original_ilocs = original_ilocs[distance_filter]
        else:
            distance_M = euclidean_distances(mesh_px, coords)
            # circle mesh type: we create spots in a hexagonal mesh
            # new_ilocs contains the indices of the columns of the sparse matrix to be created
            # original_ilocs contains the column location of the original adata.X (csr_matrix)
            new_ilocs, original_ilocs = np.nonzero(distance_M < max_distance_px)
    elif mesh_type == "hexagon":
        if optimized_binning:
            grid_x, grid_y, _extent, _last_row = _create_optimized_hex_mesh_properties(
                mesh_px
            )
            mesh_px, accum = binning_hexagon(
                coords[:, 0],
                coords[:, 1],
                gridsize=(grid_x, grid_y),
                extent=_extent,
                last_row=_last_row,
            )

            new_ilocs = np.zeros(len(coords), dtype=int)
            for i in range(len(accum)):
                new_ilocs[accum[i]] = i
        else:
            # we simply create a hex mesh, without holes. For each spot, we find the
            # hexagon it belongs to.
            # Not recommended this non-optimized approach; high memory
            # usage if spot distance is low -> O(n^2) complexity!
            # Kept for reproducibility with legacy runs of spacemake
            distance_M = euclidean_distances(mesh_px, coords)
            new_ilocs = np.argmin(distance_M, axis=0)

        original_ilocs = np.arange(new_ilocs.shape[0])
        # we need to sort the new ilocs so that they are in increasing order
        sorted_ix = np.argsort(new_ilocs)
        new_ilocs = new_ilocs[sorted_ix]
        original_ilocs = original_ilocs[sorted_ix]

    joined_coordinates = mesh_px[np.unique(new_ilocs)]

    adata.obs['n_counts'] = adata.obs['total_counts'] #np.array(adata.X.sum(axis=1))[:, 0]
    meshed_adata = aggregate_adata_by_indices(
        adata,
        idx_to_aggregate=original_ilocs,
        idx_aggregated=new_ilocs,
        coordinates_aggregated=joined_coordinates,
    )

    # set the meshing variables
    meshed_adata.uns["mesh_variables"]  = { 
        "px_by_um": px_by_um,
        "spot_diameter_um": spot_diameter_um,
        "spot_distance_um": spot_distance_um,
        "bead_diameter_um": bead_diameter_um,
        "mesh_type": mesh_type,
        "start_at_minimum": start_at_minimum
    }

    return meshed_adata



import pytest

from spacemake.map_strategy import *
from spacemake.config import ConfigFile
from spacemake.project_df import ProjectDF
from spacemake.errors import *
import os

from fixtures import configured_root, tmp_root, sm, spacemake_dir


def test_validation(configured_root):
    config = ConfigFile.from_yaml((configured_root / "config.yaml").as_posix())
    data = [
        (
            "flipped",
            "rRNA:bowtie2->genome:STAR",
            "test_hsa",
            "bowtie2:rRNA->STAR:genome",
        ),
        (
            "species_missing",
            "bowtie2:rRNA->STAR:genome",
            "test_hs",
            "<class 'spacemake.errors.ConfigVariableNotFoundError'>",
        ),
        (
            "with_cflavor",
            "bowtie2@custom_index:rRNA->STAR@default:genome",
            "test_hsa",
            "bowtie2@custom_index:rRNA->STAR@default:genome",
        ),
        (
            "unknown_cflavor",
            "rRNA:bowtie2@custom->genome:STAR@default",
            "test_hsa",
            "<class 'spacemake.errors.ConfigVariableNotFoundError'>",
        ),
        # ("flipped", "bowtie2:rRNA->STAR:genome", "rRNA:bowtie2->genome:STAR"),
    ]
    for name, mapstr, species, expect in data:
        # print(f"running test {name}")
        try:
            res = validate_mapstr(mapstr, config=config, species=species)
        except (ValueError, ConfigVariableNotFoundError) as e:
            res = str(type(e))

        print(f"test '{name}': {mapstr}-> {res} expect={expect} {expect == res}")
        assert res == expect


def test_mapstr(configured_root):
    config = ConfigFile.from_yaml((configured_root / "config.yaml").as_posix())
    data = [
        ("with_cflavor", "bowtie2@custom_index:rRNA->STAR@default:genome", None),
    ]
    for name, mapstr, expect in data:
        mr, lr = mapstr_to_targets(mapstr)
        assert mr[0].input_name == "uBAM"
        assert mr[-1].input_name == "not_rRNA.bowtie2"
        assert lr[0].link_src == "genome.STAR"
        assert lr[0].link_name == "final"


def test_get_mapped_BAM_output(configured_root):
    config = ConfigFile.from_yaml((configured_root / "config.yaml").as_posix())
    project_df = ProjectDF(
        (configured_root / "project_df.csv").as_posix(), config=config
    )

    out_files = get_mapped_BAM_output(project_df=project_df, config=config)
    print(out_files)


def test_validation_cmdline_issue_54(configured_root):
    os.chdir(configured_root.as_posix())
    data = [
        ("flipped", "rRNA:bowtie2->genome:STAR", "test_hsa", True),
        ("species_missing", "bowtie2:rRNA->STAR:genome", "test_hs", False),
        (
            "with_cflavor",
            "rRNA:bowtie2@custom_index->genome:STAR@default",
            "test_hsa",
            True,
        ),
        (
            "unknown_cflavor",
            "rRNA:bowtie2@customX->genome:STAR@defaultBLA",
            "test_hsa",
            False,
        ),
        # ("flipped", "bowtie2:rRNA->STAR:genome", "rRNA:bowtie2->genome:STAR"),
    ]
    for name, mapstr, species, expect_pass in data:
        print(f"running test {name}")
        # add
        sm(
            "projects",
            "add-sample",
            "--project-id=test",
            f"--sample-id={name}",
            f"--map-strategy={mapstr}",
            f"--R1={spacemake_dir}/test_data/reads_chr22_R1.fastq.gz",
            f"--R2={spacemake_dir}/test_data/reads_chr22_R2.fastq.gz",
            f"--species={species}",
            expect_fail=not expect_pass,
        )



import logging
import novosparc
import anndata

from spacemake.util import message_aggregation
from numpy import ndarray

logger_name = "spacemake.spatial.novosparc_integration"
logger = logging.getLogger(logger_name)


# copy of a yet-to-be-pushed novosparc function
def quantify_clusters_spatially(
    tissue: novosparc.cm.Tissue, cluster_key: str = "clusters"
) -> ndarray:
    """Maps the annotated clusters obtained from the scRNA-seq analysis onto
    the tissue space.

    :param: tissue:    the novosparc tissue object containing the gene expression data,
                      the clusters annotation and the spatial reconstruction. Assumes
                      that the cluster annotation exists in the underlying anndata object.
    :type: tissue: novosparc.cm.Tissue
    :param: cluster_key: the key of the clustering to be used for deduction
    :type: cluster_key: str
    :returns: A numpy list of cluster names, one per each spot
    :rtype: numpy.ndarray
    """
    import numpy as np

    clusters = tissue.dataset.obs[cluster_key].to_numpy().flatten()
    cluster_names = np.unique(clusters)
    ixs = np.array(
        [
            np.argmax(
                np.array(
                    [
                        np.median(
                            np.array(
                                tissue.gw[:, location][
                                    np.argwhere(clusters == cluster).flatten()
                                ]
                            )
                        )
                        for cluster in cluster_names
                    ]
                )
            )
            for location in range(len(tissue.locations))
        ]
    )
    return [cluster_names[ix] for ix in ixs]


def novosparc_denovo(
    adata: anndata.AnnData,
    num_spatial_locations: int = 5000,
    num_input_cells: int = 30000,
    locations=None,
) -> novosparc.cm.Tissue:
    """
    Given an AnnData object containing single-cell data, this function
    will try to reconstruct a 2D tissue de-novo. By default it will use
    the 500 most variable genes as markers to do this reconstruction.

    :param adata: A spacemake processed sample.
    :type adata: anndata.AnnData
    :param num_spatial_locations: The number of spatial locations to be used
        for de-novo recontruction onto a circular tissue. Ignored if locations
        are provided. If set to None this will be equal to the number of cells
        in the dataset. Default: 5000.
    :type num_spatial_locations: int
    :param num_input_cells: Number of cells from the single-cell data to be
        used for de-novo reconstruction. If set to less than the available
        number of cells, the data will be downsampled. If set to None all cells
        are used. Default: 30000
    :type num_input_cells: int
    :param locations: Numpy array of (x,y) coordinates (optional).
        Novosparc will try to reconstruct the tissue using these
        locations. If none provided, a circular tissue will be
        automatically created with num_spatial_locations number of
        locations.
    :type locations: numpy.ndarray
    :returns: A novosparc.cm.Tissue object with 2D expression information.
    :rtype: novosparc.cm.Tissue
    """
    import numpy as np
    import pandas as pd
    import scanpy as sc

    import novosparc
    import anndata

    from scipy.sparse import issparse, csc_matrix

    logger.info("Reconstructing the tissue de-novo with novosparc")

    gene_names = adata.var.index.tolist()

    num_cells, num_genes = adata.shape

    if num_input_cells is None:
        num_input_cells = num_cells

    if num_spatial_locations is None:
        num_spatial_locations = num_cells

    if num_cells > num_input_cells:
        sc.pp.subsample(adata, n_obs=num_input_cells)

    if num_cells < num_spatial_locations:
        num_spatial_locations = num_cells

    sc.pp.highly_variable_genes(adata, n_top_genes=100)
    is_var_gene = adata.var["highly_variable"]
    # select only 100 genes
    var_genes = list(is_var_gene.index[is_var_gene])

    dge_rep = adata.to_df()[var_genes]

    # if you uncomment this, adata.X will be stored not as sparse matrix, rather
    # as a regular numpy array. uncommenting this doesnt throw an error
    if issparse(adata.X):
        dense_adata = anndata.AnnData(adata.X.toarray(), obs=adata.obs, var=adata.var)
        adata = dense_adata
        del dense_adata

    if locations is None:
        # create circle locations
        locations = novosparc.gm.construct_circle(num_locations=num_spatial_locations)

    tissue = novosparc.cm.Tissue(dataset=adata, locations=locations)

    num_neighbors_s = num_neighbors_t = 5

    logger.info("Novosparc setup")
    tissue.setup_smooth_costs(
        dge_rep=dge_rep,
        num_neighbors_s=num_neighbors_s,
        num_neighbors_t=num_neighbors_t,
    )

    tissue.reconstruct(alpha_linear=0, epsilon=5e-3)

    return tissue


def novosparc_mapping(
    sc_adata: anndata.AnnData, st_adata: anndata.AnnData
) -> novosparc.cm.Tissue:
    """
    Given two AnnData objects, one single-cell and one spatial, this function
    will map the expression of the single-cell data onto the spatial data using
    shared highly variable genes as markers.

    :param sc_adata: A spacemake processed single-cell sample.
    :type sc_adata: anndata.AnnData
    :param st_adata: A spacemake processed spatial sample.
    :type st_adata: anndata.AnnData
    :returns: A novosparc.cm.Tissue object with 2D expression information.
        The locations of the Tissue will be identical to the locations of
        the spatial sample.
    :rtype: novosparc.cm.Tissue
    """
    import scanpy as sc
    import anndata
    import novosparc

    from scanpy._utils import check_nonnegative_integers
    from scipy.sparse import csc_matrix

    logger.info("Mapping single-cell data onto spatial data with novosparc")

    if check_nonnegative_integers(sc_adata.X) or check_nonnegative_integers(st_adata.X):
        # if any of the inputs is count-data, raise error
        raise SpacemakeError(
            f"External dge seems to contain raw counts. "
            + "Normalised values are expected for both sc_adata and st_adata."
        )

    # calculate top 500 variable genes for both
    sc.pp.highly_variable_genes(sc_adata, n_top_genes=500)
    sc.pp.highly_variable_genes(st_adata, n_top_genes=500)

    sc_adata_hv = sc_adata.var_names[sc_adata.var.highly_variable].to_list()
    st_adata_hv = st_adata.var_names[st_adata.var.highly_variable].to_list()

    markers = list(set(sc_adata_hv).intersection(st_adata_hv))

    logger.info(
        f"{len(markers)} number of common markers found. Using them"
        + " for reconstruction"
    )

    # save sc dge as a pandas dataframe
    dge_rep = sc_adata.to_df()[sc_adata_hv]

    if not "spatial" in st_adata.obsm:
        raise SpacemakeError(f"The object provided to st_adata is not spatial")

    locations = st_adata.obsm["spatial"]
    atlas_matrix = st_adata.to_df()[markers].values

    # make dense dataset
    dense_dataset = anndata.AnnData(
        sc_adata.X.toarray(), obs=sc_adata.obs, var=sc_adata.var
    )

    marker_ix = [dense_dataset.var.index.get_loc(marker) for marker in markers]

    tissue = novosparc.cm.Tissue(dataset=dense_dataset, locations=locations)
    num_neighbors_s = num_neighbors_t = 5

    tissue.setup_linear_cost(
        markers_to_use=marker_ix,
        atlas_matrix=atlas_matrix,
        markers_metric="minkowski",
        markers_metric_p=2,
    )
    tissue.setup_smooth_costs(
        dge_rep=dge_rep,
        num_neighbors_s=num_neighbors_s,
        num_neighbors_t=num_neighbors_t,
    )

    tissue.reconstruct(alpha_linear=0.5, epsilon=5e-3)

    return tissue


def save_novosparc_res(
    tissue: novosparc.cm.Tissue, adata_original: anndata.AnnData
) -> anndata.AnnData:
    """Save the result of novosparc spatial mapping.

    :param tissue: A Tissue object, result of the novosparc mapping.
    :type tissue: novosparc.cm.Tissue
    :param adata_original: The original AnnData which was mapped with novosparc,
        to reconstruct the Tissue. For the de-novo reconstruction, use the
        original AnnData object, when a single-cell dataset was mapped onto a
        spatial dataset, use the AnnData of the single-cell dataset.
    :type adata_original: anndata.AnnData
    :returns: An AnnData object containing the spatial information in
        .obsm['spatial'], and the imputed (after mapping) expression values
        in .X.
    :rtype: anndata.AnnData
    """
    import anndata
    import pandas as pd
    import numpy as np

    from scipy.sparse import csc_matrix

    adata_reconst = anndata.AnnData(
        csc_matrix(tissue.sdge.T), var=pd.DataFrame(index=tissue.dataset.var_names)
    )

    logger.info("Scaling mapped data to original data...")

    adata_reconst.X = (
        np.sum(adata_original.X) * adata_reconst.X / np.sum(adata_reconst.X)
    )

    adata_reconst.obsm["spatial"] = tissue.locations

    logger.info("Transferring original cluster labels...")

    for res_key in adata_original.obs.columns[
        adata_original.obs.columns.str.startswith("leiden_")
    ]:
        adata_reconst.obs[f"novosparc_{res_key}"] = quantify_clusters_spatially(
            tissue, res_key
        )

    return adata_reconst


@message_aggregation(logger_name)
def cmdline():
    """cmdline."""
    import scanpy as sc

    from spacemake.cmdline import get_novosparc_parser

    parser = get_novosparc_parser()

    args = parser.parse_args()

    sc_adata = sc.read(args.single_cell_dataset)

    if args.spatial_dataset is not None:
        # make mapping
        st_adata = sc.read(args.spatial_dataset)

        tissue_reconst = novosparc_mapping(sc_adata, st_adata)
        adata = save_novosparc_res(tissue_reconst, sc_adata)
    else:
        tissue_reconst = novosparc_denovo(sc_adata)
        adata = save_novosparc_res(tissue_reconst, sc_adata)

    adata.write(args.output)


if __name__ == "__main__":
    cmdline()



import os
import pandas as pd
import numpy as np
from glob import glob
from matplotlib.backends.backend_pdf import PdfPages
import matplotlib.pyplot as plt


def main(args):
    dfs = []
    for fname in list(args.fnames) + list(glob(args.glob_pattern)):
        print(f"loading {fname}")
        df = pd.read_csv(fname, sep='\t')
        df['stats_file'] = fname
        dfs.append(df)

    df = pd.concat(dfs)

    repriming = ['TSO,TSO_RC', 'dN-SMRT,dN-SMRT_RC', ]
    concatenation = [c for c in df.columns if c.endswith('+') and ',' not in c]
    bead = ["bead_complete", "bead_only_handle", "bead_no_dT", "bead_no_opseq"][::-1]

    # avoid crash if columns are missing
    for r in repriming + concatenation + bead:
        if r not in df.columns:
            df[r] = 0

    # print(df)
    # print(f"concat columns {concatenation}")
    # print(f"bead columns {bead}")
    df['reprimed'] = df[repriming].sum(axis=1)
    df['bead_complete'] = np.nan_to_num(df['bead_complete'], nan=0.0)
    df['concat'] = df[concatenation].sum(axis=1)
    df['bead_related'] = np.nan_to_num(df[bead].sum(axis=1), nan=0.0)
    df['bead_dropseq'] = np.nan_to_num(df['bead_no_opseq'], nan=0.0)
    df['bead_incomplete'] = df['bead_related'] - df['bead_complete'] - df['bead_dropseq']
    df['non_bead'] = 100 - df['bead_related']
    df['bead_fidelity'] = 100 * df['bead_complete'] / df['bead_related']
    df = df.fillna(0)
    # print(df)
    if args.csv_out:
        df.to_csv(args.csv_out, float_format='%.2f', sep='\t', index=False)

    def clean(txt):
        txt = os.path.basename(txt)
        t = txt\
            .replace('source/','') \
            .replace('sts_', '') \
            .replace('pb_', '') \
            .replace('ds_', '') \
            .replace('.fq', '') \
            .replace('.bam', '') \
            .replace('lima.', '')
        
        if t.count('_') > 1:
            t = "_".join(t.split('_')[:2])
        
        return t

    df['name'] = df['qfa'].apply(clean)
    # df = df.sort_values('bead_related')
    df = df.sort_values('name')

    def guess_rRNA_file(path):
        # print("guessrRNA raw path", path)
        name = os.path.basename(path).replace('.summary', '.rRNA')
        
        if args.rRNA_same_place:
            place = os.path.dirname(path)
        else:
            place = args.rRNA

        return [
            os.path.join(place, name.replace(".fq", ".txt")),
            os.path.join(place, name.replace(".fq", ".txt")).replace('.rRNA.tsv', '.txt'),
            os.path.join(place, name.replace(".fq", ".txt")).replace('.rRNA.tsv', '.rRNA.txt'),
            os.path.join(place, name.replace(".bam", ".txt").replace("lima.", "")),
            os.path.join(place, name.replace(".bam", ".txt").replace("lima.", "")).replace('.rRNA.tsv', '.txt'),
            os.path.join(place, name.replace(".bam", ".txt").replace("lima.", "")).replace('.rRNA.tsv', '.rRNA.txt'),
        ]

    rRNA_fracs = []
    for row in df[['stats_file', 'N_reads']].itertuples():
        rcount = np.nan
        for fname in guess_rRNA_file(row.stats_file):
            print(fname)
            try:
                rcount = int(open(fname).read())
            except (FileNotFoundError, ValueError):
                pass
            else:
                break
        if rcount == np.nan:
            raise ValueError

        rRNA_fracs.append(100. * rcount / row.N_reads)

    df['rRNA'] = rRNA_fracs
    # print(df[['qfa', 'rRNA']])

    def make_bars(ax, df, kinds, labels, cmap=plt.get_cmap('tab10'), w=0.9, colors=None):
        n = len(kinds)
        if colors is None:
            colors = cmap(np.linspace(0, 1, n))

        x = np.arange(len(df)) - w/2.0
        y0 = np.zeros(len(x), dtype=float)
        for kind, label, color in zip(kinds, labels, colors):
            y = np.nan_to_num(df[kind], nan=0.0)
            # print(kind)
            # print(y)
            ax.bar(x, y, bottom=y0, label=label, width=w, color=color)
            y0 += y

        ax.set_ylabel('fraction of library')
        ax.set_xticks(x)
        labels = df['name']  # [clean(fq) for fq in df['qfa']]
        ax.set_xticklabels(labels, rotation=90)
        ax.set_ylim(0, 100)

    marie = ["non_bead", "bead_incomplete", "bead_dropseq", "bead_complete", ]
    marie_colors = ["gray", "royalblue", "green", "gold"]
    
    w = max(8 / 25. * len(df), 3)
    if args.multi_page:
        pdf = PdfPages(args.breakdown)
        fig, ax1 = plt.subplots(1, figsize=(w, 4))
    else:
        fig, (ax1, ax2) = plt.subplots(2, figsize=(w, 6), sharex=True)

    make_bars(ax1, df, marie, labels=[b.replace('bead_', '') for b in marie], colors=marie_colors)
    ax1.legend(title='Marie-stats', ncol=len(marie))
    if args.multi_page:
        fig.tight_layout()
        pdf.savefig()
        plt.close()
        fig, ax2 = plt.subplots(1, figsize=(w, 4))

    make_bars(ax2, df, ["bead_fidelity"], labels=["bead fidelity"])
    ax2.set_ylabel("bead fidelity")
    if args.multi_page:
        fig.tight_layout()
        pdf.savefig()
        pdf.close()
    else:
        fig.tight_layout()
        plt.savefig(args.breakdown)

    plt.close()

    if args.multi_page:
        pdf = PdfPages(args.output)
        fig, ax1 = plt.subplots(1, figsize=(w, 4))
    else:
        fig, (ax1, ax2, ax3, ax4) = plt.subplots(4, figsize=(w, 12), sharex=True)

    # print("bead related", bead)
    make_bars(ax1, df, bead, labels=[b.replace('bead_', '') for b in bead])
    ax1.legend(title='bead-related', ncol=len(bead))
    if args.multi_page:
        fig.tight_layout()
        pdf.savefig()
        plt.close()
        fig, ax2 = plt.subplots(1, figsize=(w, 4))

    # print("repriming events", repriming)
    make_bars(ax2, df, repriming, labels=[r.split(',')[0] for r in repriming], cmap=plt.get_cmap('tab20c'))
    ax2.legend(title='repriming', ncol=len(repriming))
    if args.multi_page:
        fig.tight_layout()
        pdf.savefig()
        plt.close()
        fig, ax3 = plt.subplots(1, figsize=(w, 4))

    # print("concat events", concatenation)
    make_bars(ax3, df, concatenation, labels=concatenation, cmap=plt.get_cmap('tab20b'))
    ax3.legend(title='concatamers', ncol=len(concatenation))
    if args.multi_page:
        fig.tight_layout()
        pdf.savefig()
        plt.close()
        fig, ax4 = plt.subplots(1, figsize=(w, 4))

    make_bars(ax4, df, ["rRNA",], labels = ["rRNA"], cmap=plt.get_cmap('tab20c'))
    ax4.legend(title='human rRNA', ncol=1)
    if args.multi_page:
        fig.tight_layout()
        pdf.savefig()
        pdf.close()
    else:
        fig.tight_layout()
        plt.savefig(args.output)

    plt.close()


def setup_parser(parser):
    parser.add_argument("fnames", nargs='*')
    parser.add_argument("--output", default="pb_overview.pdf",
                        help="path/name of detailed report PDF")
    parser.add_argument("--csv-out", default="all_pb_stats.csv",
                        help="path/name of detailed report PDF")
    parser.add_argument("--breakdown", default="bead_overview.pdf",
                        help="path/name of bead report (Marie style) PDF")
    parser.add_argument("--glob-pattern", default="stats/*summary.tsv",
                        help="search pattern to gather summary files generated by the scan command")
    parser.add_argument("--rRNA", default="rRNA/", 
                        help="path to search for rRNA counts corresponding to samples")
    parser.add_argument("--rRNA-same-place", default=False, action='store_true',
                        help="If set, look for rRNA txt file with same sample name in same directory")
    parser.add_argument("--multi-page", default=False, action="store_true",
                        help="If set, generate multiple PDF pages instead of subplots")


if __name__ == "__main__":
    # setup own parser
    import argparse
    parser = argparse.ArgumentParser(prog='pb_overview')
    setup_parser(parser)
    main(parser.parse_args())



#!/usr/bin/env python3
__version__ = "0.9"
__author__ = ["Marvin Jens"]
__license__ = "GPL"
__email__ = ["marvin.jens@mdc-berlin.de"]

from spacemake.preprocess.fastq import (
    parse_args,
    setup_logging,
    main_combinatorial,
    main_dropseq,
)

from spacemake.parallel import ExceptionLogging


def cmdline():
    with ExceptionLogging("main"):
        args = parse_args()
        NO_CALL = args.na
        setup_logging(args)

        if args.out_format == "bam" and not args.read2:
            raise ValueError("bam output format requires --read2 parameter")

        if ("bc1" in args.cell and not args.bc1_ref) or (
            "bc2" in args.cell and not args.bc2_ref
        ):
            raise ValueError(
                "bc1/2 are referenced in --cell or --cell-raw, but no reference barcodes are specified via --bc{{1,2}}-ref"
            )

        if args.bc1_ref or args.bc2_ref:
            main_combinatorial(args)
        else:
            main_dropseq(args)


if __name__ == "__main__":
    cmdline()



import os
import subprocess
import sys
import time
import yaml

from spacemake.contrib import __version__
from spacemake.project_df import get_global_ProjectDF
from spacemake.util import sync_timestamps, sync_symlink_mtime

def find_bam_files(folder):
    """
    Finds all .bam files in the given folder and checks if any of them is a symlink.
    
    Returns a list of tuples of type (str, bool), e.g. ('bam_file', False)
    """
    if not os.path.isdir(folder):
        raise ValueError(f"The provided path {folder} is not a valid directory.")
    
    # Find files and check for symlinks
    bam_files = [f for f in os.listdir(folder) if f.endswith('.bam')]
    bam_file_paths = [os.path.join(folder, f) for f in bam_files]
    are_symlinks = [os.path.islink(bam_file_path) for bam_file_path in bam_file_paths]

    return list(zip(bam_file_paths, are_symlinks))


def get_map_strategy_sequences(project_id, sample_id):
    """
    Returns a dictionary of reference_types and their location, e.g. {rRNA : /path/to/disk/sequence.fa}
    """
    pdf = get_global_ProjectDF()

    species = pdf.get_sample_info(project_id, sample_id)['species']
    map_strategy = pdf.get_sample_info(project_id, sample_id)['map_strategy']
    sequence_type = [mapping.split(':')[1] for mapping in map_strategy.split('->')]

    reference_type = {st : f"species_data/{species}/{st}/sequence.fa" for st in sequence_type}

    return reference_type


def check_if_sample_is_processed(project_id, sample_id):
    """
    Checks if the sample is newly added and spacemake has not been run yet.
    Most likely to happen when the sample is added with a newer version of
    spacemake compared to the existing project_df.

    Returns:
        bool: True if sample has been already processed.
    """
    sample_folder = os.path.join('projects', project_id, 'processed_data',
                                  sample_id, 'illumina', 'complete_data')
    if not os.path.isdir(sample_folder):
        return False
    else:
        return len(os.listdir(sample_folder)) > 1

def check_if_genome_files_are_on_disk(project_id, sample_id):
    """
    Checks if tha genome fasta files exist on disk. Raises an error if they don't
    so that the snakemake logic won't continue.
    """
    species_sequences = get_map_strategy_sequences(project_id, sample_id)

    for file in species_sequences.values():
        if not os.path.exists(file):
            raise RuntimeError(f"Required genome file missing: {file}")

def check_if_all_files_exist(project_id, sample_id, file_type):
    """
    Checks if all required files exist for the map_strategy of the sample.

    Args:
        file_type (str): The type of file to check. Must be either 'bam' or 'cram'

    Returns:
        bool: True if all required files exist. False otherwise.
    """
    project_folder = os.path.join('projects', project_id, 'processed_data',
                                  sample_id, 'illumina', 'complete_data')

    pdf = get_global_ProjectDF()
    sample_is_merged = pdf.get_sample_info(project_id, sample_id)['is_merged']
    map_strategy = pdf.get_sample_info(project_id, sample_id)['map_strategy']

    aligner = [mapping.split(':')[0] for mapping in map_strategy.split('->')]
    sequence_type = [mapping.split(':')[1] for mapping in map_strategy.split('->')]

    files_expected = [f"{x}.{y}.{file_type}" for x, y in zip(sequence_type, aligner)]

    if sample_is_merged:
        files_expected = ['final.polyA_adapter_trimmed.merged.' + file_type]

    all_files_exist = True
    for file in files_expected:
        file_path = os.path.join(project_folder, file)
        if not os.path.exists(file_path) or os.path.getsize(file_path) == 0:
            all_files_exist = False
            print('file', file_path, 'does not exist or is empty.')
            break

    return all_files_exist


def convert_bam_to_cram(project_id, sample_id, threads=4):
    """
    Converts all BAM files to CRAM and updates the timestamps to those of the
    original files. Symbolic links are treated as such.
    """
    pdf = get_global_ProjectDF()
    sample_is_merged = pdf.get_sample_info(project_id, sample_id)['is_merged']
    
    species_sequences = get_map_strategy_sequences(project_id, sample_id)

    if check_if_all_files_exist(project_id, sample_id, 'bam'):
        print('All required BAM files exist on disk -- will continue with CRAM conversion.')
    else:
        print('Not all required BAM files were found on disk -- will not continue with CRAM conversion.')
        return

    project_folder = os.path.join('projects', project_id, 'processed_data',
                                  sample_id, 'illumina', 'complete_data')    
    bam_files = find_bam_files(project_folder)
    bam_files = sorted(bam_files, key=lambda x: ("final" not in x[0], x[1])) # sort to make sure final is always checked first
    
    # simple dictionary to keep track of which ref_type is final
    ref_type_final = {}
    
    for idx in range(len(bam_files)):
        bam_filename, bam_file_is_symlink = bam_files[idx]
        bam_filename_prefix = bam_filename.rsplit('.', 1)[0]
        cram_filename = bam_filename_prefix + '.cram'

        if os.path.exists(cram_filename) and os.path.getsize(cram_filename) > 0:
            print('CRAM file', cram_filename, 'already exists. Skipping conversion.')
            continue
        elif os.path.exists(cram_filename) and os.path.getsize(cram_filename) == 0:
            print('WARNING: CRAM file', cram_filename, 'already exists, but file size is 0. Possible permission error.')

        # Special case for now
        if 'unaligned' in bam_filename:
            cram_filename = os.path.join(project_folder, 'unaligned_bc_tagged.polyA_adapter_trimmed.cram')
            if os.path.exists(cram_filename) and os.path.getsize(cram_filename) > 0:
                print('CRAM file', cram_filename, 'already exists. Skipping conversion.')
                continue
            else:
                print('Converting', bam_filename, 'to', cram_filename,
                    '...', time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()))
                subprocess.run(
                    [
                        "samtools", "view",
                        "-T", species_sequences['genome'],
                        "-C",
                        "--threads", str(threads),
                        "-o", cram_filename,
                        bam_filename
                    ])
                sync_timestamps(bam_filename, cram_filename)
                continue
        
        if bam_file_is_symlink:
            true_bam_filename = os.readlink(bam_filename)
            true_bam_filename_prefix = true_bam_filename.rsplit('.', 1)[0]

            # Keep track of what's final to split the not_ files later.
            if 'final' in bam_filename:
                ref_type_final[true_bam_filename.rsplit('.')[0]] = True

            try:
                os.symlink(true_bam_filename_prefix + '.cram', cram_filename)
                sync_symlink_mtime(bam_filename, cram_filename)

            except FileExistsError:
                print('CRAM symlink already exists.')

        else:
            print('Converting', bam_filename, 'to', cram_filename, 
            '...', time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()))

            for ref_type in species_sequences:
                if sample_is_merged:
                    ref_sequence = species_sequences['genome']
                    break
                if ref_type in bam_filename:
                    ref_sequence = species_sequences[ref_type]
                    break
            
            # warn if reference is gzip-compressed
            if ref_sequence.endswith('.gz'):
                print(f"ERROR: Reference file '{ref_sequence}' is gzip-compressed (.gz), which samtools cannot use for CRAM conversion.", file=sys.stderr)
                print("Please use an uncompressed FASTA (.fa) or bgzip-compressed version with proper indexing (.fai and .gzi).", file=sys.stderr)
                sys.exit(1)

            if ref_type in ref_type_final:
                #split files appropriately
                subprocess.run(
                    [
                        "samtools", "view",
                        "-T", ref_sequence,
                        "-C",
                        "-F 4",
                        "--threads", str(threads),
                        "-o", cram_filename,
                        bam_filename
                    ])
                
                directory, filename = os.path.split(cram_filename)
                not_filename = 'not_' + filename

                subprocess.run(
                    [
                        "samtools", "view",
                        "-T", ref_sequence,
                        "-C",
                        "-f 4",
                        "--threads", str(threads),
                        "-o", os.path.join(directory, not_filename),
                        bam_filename
                    ])
                
                sync_timestamps(bam_filename, os.path.join(directory, not_filename))

            else:
                subprocess.run(
                    [
                        "samtools", "view",
                        "-T", ref_sequence,
                        "-C",
                        "-F 4",
                        "--threads", str(threads),
                        "-o", cram_filename,
                        bam_filename
                    ])

                if sample_is_merged:
                    subprocess.run(
                        [
                            "touch", os.path.join(project_folder, 'final.polyA_adapter_trimmed.cram')
                        ])
                
            sync_timestamps(bam_filename, cram_filename)


def rename_log_files(project_id, sample_id):
    """
    Rename any .bam.log files (created by bowtie) so that qc_sheets and other reports
    generation downstream does not fail.
    """
    project_folder = os.path.join('projects', project_id, 'processed_data',
                                  sample_id, 'illumina', 'complete_data')    
    
    log_files = [f for f in os.listdir(project_folder) if f.endswith('.bam.log')]

    for log_filename in log_files:
        log_filename_prefix = log_filename.rsplit('.', 2)[0]
        new_log_filename = log_filename_prefix + '.cram.log'
        os.rename(os.path.join(project_folder, log_filename),
                  os.path.join(project_folder, new_log_filename))


def remove_bam_files(project_folder, output_file_path):
    bam_files = find_bam_files(project_folder)

    cram_files = [
        os.path.join(project_folder, f)
        for f in os.listdir(project_folder)
        if f.endswith(".cram") and os.path.isfile(os.path.join(project_folder, f))
    ]

    # make sure the generated CRAMs are not empty -- otherwise raise an error and do not remove any BAMs
    for cram_file in cram_files:
        if os.path.getsize(cram_file) == 0:
            raise ValueError(f"CRAM file is empty: {cram_file}. Aborting BAM file removal.")

    total_bam_size = sum(
        os.path.getsize(bam[0]) for bam in bam_files if os.path.exists(bam[0])
    )
    total_cram_size = sum(
        os.path.getsize(cram) for cram in cram_files if os.path.exists(cram)
    )

    # remove BAM files
    deleted_files = []
    for bam in bam_files:
        if os.path.exists(bam[0]):
            os.remove(bam[0])
            deleted_files.append(bam[0])

    saved_bytes = total_bam_size - total_cram_size
    saved_gb = saved_bytes / (1024 ** 3)

    # write report
    with open(output_file_path, "w") as out:
        out.write(f"BAM files removed: {len(deleted_files)}\n")
        out.write("Files:\n")
        for f in deleted_files:
            out.write(f"  - {f}\n")
        out.write(f"Total disk space saved: {saved_gb:.2f} GB\n")

    print(f"Deleted {len(deleted_files)} BAM files, saved ~{saved_gb:.2f} GB")


def update_adapters_in_config():
    """
    Update the adapters seciton of the config to current version.
    """
    initial_config = os.path.join(os.path.dirname(__file__), "data/config/config.yaml")

    with open("config.yaml", "r") as f1, open(initial_config, "r") as f2:
        config = yaml.safe_load(f1)
        config_latest = yaml.safe_load(f2)

    outdated = (
        "adapter_flavors" not in config or
        config.get("adapter_flavors") != config_latest.get("adapter_flavors")
    )

    if outdated:
        print("Outdated config.yaml was identified. Migrating to the latest version.")
        config["adapter_flavors"] = config_latest["adapter_flavors"]
        with open("config.yaml", "w") as f:
            yaml.dump(config, f)


def update_version_in_config():
    """
    Starting with v0.9, the spacemake version is recorded inside the config.yaml
    as 'spacemake_version: '.

    For migration of earlier 0.8x -> 0.9, this entry has to be added.
    For future mirgations, the record is updated. 
    """
    # save original timestamps
    try:
        original_stat = os.stat("config.yaml")
    except FileNotFoundError:
        print("config.yaml not found.")
        return
    
    with open("config.yaml", "r") as f:
        config = yaml.safe_load(f)

    if "spacemake_version" not in config:
        config["spacemake_version"] = __version__
        with open("config.yaml", "w") as f:
            yaml.dump(config, f)

        # restore timestamps
        os.utime("config.yaml", (original_stat.st_atime, original_stat.st_mtime))

    else:
        # placeholder for future migrations
        return



import pandas as pd
import numpy as np
import os
import ncls
import argparse
import re
import gzip
import logging
import ncls
import itertools
import pickle
from time import time
from collections import defaultdict, OrderedDict, deque

logging.basicConfig(level=logging.DEBUG)


def attr_to_dict(attr_str):
    """
    Helper for reading the GTF. Parsers the attr_str found in the GTF
    attribute column (col 9). Returns a dictionary from the key, value
    pairs.
    """
    d = {}
    for match in re.finditer(r"(\w+) \"(\S+)\";", attr_str):
        key, value = match.groups()
        d[key] = value

    return d


def load_GTF(
    src,
    attributes=["gene_id", "gene_type", "gene_name"],
    features=["exon", "CDS", "UTR"],
):
    if type(src) is str:
        if src.endswith(".gz"):
            src = gzip.open(src, "rt")
        else:
            src = open(src, "rt")

    fset = set(features)
    data = []
    for line in src:
        parts = line.split("\t")
        if len(parts) != 9:
            # not a GFF formatted line
            continue

        chrom, source, feature, start, end, score, strand, frame, attr_str = parts
        if feature not in fset:
            continue

        attrs = attr_to_dict(attr_str)
        row = [chrom, feature, int(start) - 1, int(end), strand] + [
            attrs[a] for a in attributes
        ]
        data.append(row)

    return pd.DataFrame(
        data, columns=["chrom", "feature", "start", "end", "strand"] + attributes
    ).drop_duplicates()


## Annotation classification matrix
default_map = {
    # CDS   UTR    exon -> ga gf tag values
    (True, False, True): "CDS",  # CODING
    (True, True, True): "CDS",  # CODING
    (False, True, True): "UTR",  # UTR
    (False, False, True): "non-coding",  # CODING
    (True, False, False): "intron",  # INTRONIC
    (False, True, False): "intron",  # INTRONIC
    (False, False, False): "intron",  # INTRONIC
}


class GTFClassifier:
    def __init__(
        self,
        df,
        hierarchy=["CDS", "UTR", "non-coding", "intron"],
        feature_map=default_map,
    ):
        self.feature_map = feature_map
        self.hierarchy = hierarchy
        self.prio = {}
        for i, h in enumerate(hierarchy):
            self.prio[h] = i

        feat2idx = {
            "CDS": 0,
            "UTR": 1,
            "exon": 2,
            # "transcript": 3,
        }
        df["feature_idx"] = df["feature"].apply(lambda x: feat2idx[x])
        self.df = df.values

    def process(self, ids):
        gene_features = defaultdict(lambda: np.zeros(3, dtype=bool))
        gene_names = {}
        gene_types = {}
        # iterate over the overlapping GTF features only once.
        # sort by gene_id as we go
        for i in ids:
            row = self.df[i][5:]
            # print(row)
            (strand, gene_id, gene_type, gene_name, feature_idx) = row
            gene_names[gene_id] = gene_name
            gene_types[gene_id] = gene_type
            # feature_idx encodes CDS, UTR, exon, transcript records
            # as numbers to quickly construct a correct absence/presence
            # boolean flag-vector to be used for the lookup
            gene_features[gene_id][feature_idx] = True

        top_prio = np.inf
        gn_by_prio = defaultdict(list)
        gf_by_prio = defaultdict(list)
        gt_by_prio = defaultdict(list)
        gs_by_prio = defaultdict(list)

        for gene_id, feature_flags in gene_features.items():
            # here the boolean vector is converted to a tuple
            # which can serve as a key for lookup
            if not feature_flags[-1]:
                print(gene_id)
                for i in ids:
                    print(self.df[i])

            cls = self.feature_map[tuple(feature_flags)]
            # print(tx_id, gene, gene_id, gene_type, features, cls)
            prio = self.prio[cls]
            top_prio = min(top_prio, prio)
            gn_by_prio[prio].append(gene_names[gene_id])
            gf_by_prio[prio].append(cls)
            gt_by_prio[prio].append(gene_types[gene_id])
            gs_by_prio[prio].append(strand)

        return (
            gf_by_prio[top_prio],
            gn_by_prio[top_prio],
            gs_by_prio[top_prio],
            gt_by_prio[top_prio],
        )


class CompiledClassifier:
    def __init__(self, df, classifications):
        self.cid_table = df["cid"].values
        self.classifications = np.array(classifications, dtype=object)

    @staticmethod
    def get_filenames(path):
        cdf_path = os.path.join(path, "non_overlapping.csv")
        cclass_path = os.path.join(path, "classifications.npy")

        return cdf_path, cclass_path

    @staticmethod
    def files_exist(path):
        cdf_path, cclass_path = CompiledClassifier.get_filenames(path)
        return os.access(cdf_path, os.R_OK) and os.access(cclass_path, os.R_OK)

    # The only piece of work left is merging multiple pre-classified annotations
    def joiner(self, cidx):
        # we want to report each combination only once
        # Note: This code path is executed rarely
        seen = set()
        gf = []
        gn = []
        gs = []
        gt = []
        for i in tuple(cidx):
            cid = self.cid_table[i]
            for ann in zip(*self.classifications[cid]):
                if not ann in seen:
                    seen.add(ann)
                    gf.append(ann[0])
                    gn.append(ann[1])
                    gs.append(ann[2])
                    gt.append(ann[3])

        return gf, gn, gs, gt

    def process(self, cidx):
        # fast path
        if len(cidx) == 1:
            # fast path
            idx = tuple(cidx)[0]
            return self.classifications[self.cid_table[idx]]

        return self.joiner(cidx)


## Helper functions for working with NCLS
def query(nc, x0, x1):
    """
    report only the target_ids from the overlapping
    (start, end, id) tuples returned by NCLS.find_overlap().
    format is frozenset bc that can be hashed and used as a key.
    """
    return frozenset((o[2] for o in nc.find_overlap(x0, x1)))


def decompose(nc):
    """
    Iterate over the unique start or end coordinates of all the intervals
    in a NCLS, essentially decomposing the original NCLS into non-overlapping
    segments of unique target_id combinations.
    This is used to *compile* an NCLS. Compiled, here means that target_ids
    reference not original GTF features, but enumerated, unique combinations of
    the GTF features as they are positioned and potentially overlap in the
    reference sequence.
    The advantage is that each such combination has to be processed into the
    tagging information only once, and that this processing can be done prior
    to any tagging, reducing the complexity to a simple array lookup with O(1).

    :param nc: The NCLS (nested list) with original genomic features from GTF
    :type nc: NCLS64
    :yield: (start, end, frozenset(target_ids) )
    :rtype: None
    """
    logger = logging.getLogger("compile")
    logger.info(f"compiling strand into non-overlapping and pre-classified annotations")

    starts, ends, ids = np.array(nc.intervals()).T
    logger.debug(f"retrieved {len(starts)} intervals from ncls")

    breakpoints = np.array(sorted(set(starts) | set(ends)))
    logger.debug(f"identified {len(breakpoints)} breakpoints")

    if not len(starts):
        return

    # since we sorted, the first must be a start position
    last_pos = breakpoints[0]
    last_key = query(nc, last_pos, last_pos + 1)
    for bp in breakpoints[1:]:
        # depending on wether this is another start or
        # end position, the transition point can be off by one nt
        # in either direction. Most robust fix was to scan all 3
        # options until the key changes (while ensuring that end > start)
        for x in [-1, 0, +1]:
            if bp + x <= last_pos:
                continue

            key = query(nc, bp + x, bp + x + 1)
            if key != last_key:
                if len(last_key):
                    # ensure we do not yield the empty set
                    yield last_pos, bp + x, last_key
                last_key = key
                last_pos = bp + x
                break


class GenomeAnnotation:
    """
    NCLS-based lookup of genome annotation
    """

    logger = logging.getLogger("GenomeAnnotation")

    def __init__(self, df, processor, is_compiled=False):
        """
        [summary]

        :param df: Annotation data of either raw GTF features or pre-classified combinations
        :type df: DataFrame with at least the following columns: chrom, strand, start, end
        :param processor: evaluates indices into df and returns annotation
            the indices point into the dataframe to all
            overlapping features
        :type processor: function that takes frozenset(indices) as sole argument
        """
        # self.df = df
        self.processor = processor

        # find all unique combinations of chrom + strand
        strands = df[["chrom", "strand"]].drop_duplicates().values
        self.strand_keys = [tuple(s) for s in strands]
        self.strand_map = {}
        self.empty = frozenset([])

        t0 = time()
        for strand_key in sorted(self.strand_keys):
            chrom, strand = strand_key

            d = df.query(f"chrom == '{chrom}' and strand == '{strand}'")
            nested_list = ncls.NCLS(d["start"], d["end"], d.index)
            self.strand_map[strand_key] = nested_list

        dt = time() - t0
        self.logger.info(
            f"constructed nested lists of {len(df)} features on {len(self.strand_keys)} strands in {dt:.3f}s"
        )
        self.is_compiled = is_compiled

    @classmethod
    def from_compiled_index(cls, path):
        cdf_path, cclass_path = CompiledClassifier.get_filenames(path)
        t0 = time()
        cdf = pd.read_csv(cdf_path, sep="\t")
        classifications = np.load(cclass_path, allow_pickle=True)
        dt = time() - t0
        cls.logger.info(
            f"re-read compiled annotation index with {len(cdf)} original GTF feature combinations and {len(classifications)} in {dt:.3f} seconds"
        )
        ## Create a secondary Annotator which uses the non-overlapping combinations
        ## and the pre-classified annotations for the actual tagging
        cl = CompiledClassifier(cdf, classifications)
        gc = cls(cdf, lambda idx: cl.process(idx), is_compiled=True)
        return gc

    @classmethod
    def from_GTF(cls, gtf, df_cache=""):
        # load GTF the first time. Need to build compiled annotation
        t0 = time()
        df = load_GTF(gtf)
        dt = time() - t0
        cls.logger.info(f"loaded {len(df)} GTF records in {dt:.3f} seconds")
        if df_cache:
            df.to_csv(df_cache, sep="\t")

        ## Build NCLS with original GTF features
        cl = GTFClassifier(df)
        ga = cls(df, lambda idx: cl.process(idx))
        return ga

    @classmethod
    def from_uncompiled_df(cls, path):
        t0 = time()
        df = pd.read_csv(path, sep="\t")
        dt = time() - t0
        cls.logger.info(f"loaded {len(df)} tabular records in {dt:.3f} seconds")

        ## Build NCLS with original GTF features
        cl = GTFClassifier(df)
        ga = cls(df, lambda idx: cl.process(idx))
        return ga

    def sanity_check(self, df):
        for strand, nc in self.strand_map.items():
            starts, ends, idx = np.array(nc.intervals()).T
            self.logger.debug(
                f"checking {strand} with starts from {starts.min()}-{starts.max()} and idx from {idx.min()}-{idx.max()}"
            )

            assert idx.max() < len(df)

    def query_idx(self, chrom, start, end, strand):
        strand_key = (chrom, strand)
        if not strand_key in self.strand_map:
            return self.empty

        nested_list = self.strand_map[strand_key]
        return query(nested_list, start, end)

    def query_idx_blocks(self, chrom, strand, blocks):
        idx = set()
        for start, end in blocks:
            idx |= self.query_idx(chrom, start, end, strand)
        return frozenset(idx)

    def query(self, chrom, start, end, strand):
        idx = self.query_idx(chrom, start, end, strand)
        return self.processor(idx)

    def query_blocks(self, chrom, strand, blocks):
        idx = self.query_idx_blocks(chrom, strand, blocks)
        return self.processor(idx)

    def compile(self, path=""):
        chroms = []
        strands = []
        cstarts = []
        cends = []
        cids = []
        cidx = []
        cid_lkup = {}

        t0 = time()
        for strand_key, nc in self.strand_map.items():
            chrom, strand = strand_key
            self.logger.info(f"decomposing {chrom} {strand}")
            for start, end, idx in decompose(nc):
                # print(f"start={start} end={end} idx={idx}")
                chroms.append(chrom)
                strands.append(strand)
                cstarts.append(start)
                cends.append(end)
                if idx not in cid_lkup:
                    cid_lkup[idx] = len(cid_lkup)
                    cidx.append(idx)

                cids.append(cid_lkup[idx])

        self.logger.info("done")
        cdf = pd.DataFrame(
            dict(chrom=chroms, strand=strands, start=cstarts, end=cends, cid=cids)
        )
        dt = time() - t0
        self.logger.debug(f"decomposed original GTF annotation in {dt:.3f} seconds")

        # pre-classify all unique feature combinations
        classifications = []
        t0 = time()
        for n, idx in enumerate(cidx):
            res = self.processor(idx)
            classifications.append(res)

        ## Turn into np.array to turn lookup into super-fast array index operation
        classifications = np.array(classifications, dtype=object)
        dt = time() - t0
        self.logger.debug(
            f"pre-classified {n} feature combinations in {dt:.3f} seconds"
        )

        if path:
            cdf_path, cclass_path = CompiledClassifier.get_filenames(path)
            ## Store the compiled DataFrame and pre-classifications
            t0 = time()
            cdf.to_csv(cdf_path, sep="\t", index=False)
            np.save(cclass_path, classifications)
            dt = time() - t0
            self.logger.debug(f"stored compiled annotation index in {dt:.3f} seconds")

        ## Create a secondary Annotator which uses the non-overlapping combinations
        ## and the pre-classified annotations for the actual tagging
        cl = CompiledClassifier(cdf, classifications)
        gc = GenomeAnnotation(cdf, lambda idx: cl.process(idx), is_compiled=True)

        return gc

    def annotate_BAM(self, src, out, antisense=False, interval=5):
        import pysam

        as_strand = {"+": "-", "-": "+"}
        self.logger.info(
            f"beginning BAM annotation: {src} -> {out}. is_compiled={self.is_compiled}"
        )
        bam = pysam.AlignmentFile(src)
        out = pysam.AlignmentFile(out, "wbu", template=bam)
        t0 = time()
        T = interval
        for n, read in enumerate(bam.fetch(until_eof=True)):
            if not read.is_unmapped:
                chrom = bam.get_reference_name(read.tid)
                strand = "-" if read.is_reverse else "+"

                gf, gn, gs, gt = self.query_blocks(chrom, strand, read.get_blocks())
                if antisense:
                    gf_as, gn_as, gs_as, gt_as = self.query_blocks(
                        chrom, as_strand[strand], read.get_blocks()
                    )
                    gf += gf_as
                    gn += gn_as
                    gs += gs_as
                    gt += gt_as

                if len(gf):
                    read.tags += [
                        ("gF", ",".join(gf)),
                        ("gN", ",".join(gn)),
                        ("gS", ",".join(gs)),
                        ("gT", ",".join(gt)),
                    ]
                else:
                    read.tags += [("gF", "INTERGENIC")]

            out.write(read)
            dt = time() - t0
            if dt > T:
                self.logger.info(
                    f"processed {n} alignments in {dt:.2f} seconds ({n/dt:.2f} reads/second)"
                )
                T += interval

        self.logger.info(
            f"processed {n} alignments in {dt:.2f} seconds ({n/dt:.2f} reads/second)"
        )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--gtf",
        required=True,
        help="path to the original annotation (e.g. gencodev38.gtf.gz)",
    )
    parser.add_argument(
        "--tabular",
        default="",
        help="path to tabular version of the relevant features only (e.g. gencodev38.tsv)",
    )
    parser.add_argument(
        "--compiled",
        help="path to keep a compiled version of the GTF in (used as cache)",
    )
    parser.add_argument(
        "--use-compiled",
        default=False,
        action="store_true",
        help="enable using the compiled version of GenomeAnnotation (faster)",
    )
    parser.add_argument(
        "--antisense",
        default=False,
        action="store_true",
        help="enable annotating against the opposite strand (antisense to the alignment) as well",
    )
    parser.add_argument(
        "--dropseqtools",
        default=False,
        action="store_true",
        help="emulate dropseqtools behavior",
    )

    parser.add_argument("--bam-in", help="path for the input BAM to be tagged")
    parser.add_argument(
        "--bam-out",
        help="path for the tagged BAM output",
    )
    args = parser.parse_args()

    if CompiledClassifier.files_exist(args.compiled) and args.use_compiled:
        ga = GenomeAnnotation.from_compiled_index(args.compiled)

    elif args.tabular and os.access(args.tabular, os.R_OK):
        ga = GenomeAnnotation.from_uncompiled_df(args.tabular)

    else:
        ga = GenomeAnnotation.from_GTF(args.gtf, df_cache=args.tabular)

    # perform compilation if that's what we want
    if not ga.is_compiled and args.use_compiled:
        ga = ga.compile(args.compiled)

    ga.annotate_BAM(args.bam_in, args.bam_out, antisense=args.antisense)



import argparse
import os
import logging
from spacemake.util import message_aggregation
from spacemake.config import ConfigFile
from spacemake.contrib import __version__, __license__, __author__, __email__
import yaml

logger_name = "spacemake.main"
logger = logging.getLogger(logger_name)


def get_project_sample_parser(allow_multiple=False, prepend="", help_extra=""):
    """
    Return a parser for project_id's and sample_id's

    :param allow_multiple: if true, we allow multiple projects and samples,
        and the parser will have a `--project_id_list` and a `--sample_id_list`
        parameter
    :param prepend: a string that will be prepended before the arguments
    :param help_extra: extra help message
    :return: a parser object
    :rtype: argparse.ArgumentParser
    """
    logger.debug(f"get_project_sample_parser(prepend={prepend}) called")

    parser = argparse.ArgumentParser(allow_abbrev=False, add_help=False)
    project_argument = "project_id"
    sample_argument = "sample_id"
    required = True
    nargs = None
    default = None

    if allow_multiple:
        project_argument = f"{project_argument}_list"
        sample_argument = f"{sample_argument}_list"
        required = False
        nargs = "*"
        default = []

    parser.add_argument(
        f"--{prepend}{project_argument.replace('_', '-')}",
        type=str,
        required=required,
        nargs=nargs,
        default=default,
        help=f"{project_argument} {help_extra}",
        dest=f"{prepend.replace('-', '_')}{project_argument}",
    )
    parser.add_argument(
        f"--{prepend}{sample_argument.replace('_', '-')}",
        type=str,
        required=required,
        nargs=nargs,
        default=default,
        help=f"{sample_argument} {help_extra}",
        dest=f"{prepend.replace('-', '_')}{sample_argument}",
    )

    # Add legacy snake_case arguments
    parser.add_argument(
        f"--{prepend}{project_argument}",
        type=str,
        required=False,  # Not required since kebab-case takes precedence
        nargs=nargs,
        default=None,  # No default here; rely on kebab-case
        dest=f"{prepend.replace('-', '_')}{project_argument}",
    )
    parser.add_argument(
        f"--{prepend}{sample_argument}",
        type=str,
        required=False,
        nargs=nargs,
        default=None,
        dest=f"{prepend.replace('-', '_')}{sample_argument}",
    )

    return parser


def get_add_sample_sheet_parser():
    """
    Returns parser for sample sheet addition

    :return: parser
    :rtype: argparse.ArgumentParser
    """
    logger.debug("get_add_sample_sheet_parser() called")
    parser = argparse.ArgumentParser(
        allow_abbrev=False,
        description="add a new sample sheet to the samples",
        add_help=False,
    )

    parser.add_argument(
        "--sample-sheet",
        type=str,
        help="the path to the Illumina sample sheet",
        required=True,
        dest="sample_sheet",
    )
    parser.add_argument(
        "--sample_sheet",
        type=str,
        help=argparse.SUPPRESS,
        required=False,
        dest="sample_sheet",
    )

    parser.add_argument(
        "--basecalls-dir",
        type=str,
        help="path to the basecalls directory",
        required=True,
        dest="basecalls_dir",
    )
    parser.add_argument(
        "--basecalls_dir",
        type=str,
        help=argparse.SUPPRESS,
        required=False,
        dest="basecalls_dir",
    )

    return parser


def get_sample_main_variables_parser(
    defaults=False,
    species_required=False,
    main_variables=[
        "barcode_flavor",
        "adapter_flavor",
        "species",
        "puck",
        "run_mode",
        "map_strategy",
    ],
):
    parser = argparse.ArgumentParser(allow_abbrev=False, add_help=False)
    logger.debug(
        f"get_sample_main_variables_parser called with main_variables={main_variables}"
    )

    if "barcode_flavor" in main_variables:
        parser.add_argument(
            "--barcode-flavor",
            type=str,
            default="default" if defaults else None,
            help="barcode flavor for this sample",
            dest="barcode_flavor",
        )
        parser.add_argument(
            "--barcode_flavor",
            type=str,
            default="default" if defaults else None,
            help=argparse.SUPPRESS,
            dest="barcode_flavor",
        )

    if "adapter_flavor" in main_variables:
        parser.add_argument(
            "--adapter-flavor",
            type=str,
            default="default" if defaults else None,
            help="barcode flavor for this sample",
            dest="adapter_flavor",
        )
        parser.add_argument(
            "--adapter_flavor",
            type=str,
            default="default" if defaults else None,
            help=argparse.SUPPRESS,
            dest="adapter_flavor",
        )

    if "species" in main_variables:
        parser.add_argument(
            "--species",
            type=str,
            help="add the name of the species for this sample",
            required=species_required,
        )

    if "map_strategy" in main_variables:
        parser.add_argument(
            "--map-strategy",
            type=str,
            help="string constant definining mapping strategy. Can be multi-stage and use bowtie2 or STAR (see documentation)",
            required=False,
            default="STAR:genome",
            dest="map_strategy",
        )
        parser.add_argument(
            "--map_strategy",
            type=str,
            help=argparse.SUPPRESS,
            required=False,
            default=None,
            dest="map_strategy",
        )

    if "puck" in main_variables:
        parser.add_argument(
            "--puck",
            type=str,
            help="name of the puck for this sample. if puck_type contains a \n"
            + "`barcodes` path to a coordinate file, those coordinates\n"
            + " will be used when processing this sample. if \n"
            + " not provided, a default puck will be used with \n"
            + "width_um=3000, spot_diameter_um=10",
        )

        parser.add_argument(
            "--puck-barcode-file-id",
            type=str,
            help="puck_barcode_file_id of the sample to be added/update",
            nargs="+",
            dest="puck_barcode_file_id",
        )
        parser.add_argument(
            "--puck_barcode_file_id",
            type=str,
            help=argparse.SUPPRESS,
            nargs="+",
            dest="puck_barcode_file_id",
        )

    parser.add_argument(
        "--puck-barcode-file",
        type=str,
        nargs="+",
        help="the path to the file contining (x,y) positions of the barcodes",
        dest="puck_barcode_file",
    )
    parser.add_argument(
        "--puck_barcode_file",
        type=str,
        nargs="+",
        help=argparse.SUPPRESS,
        dest="puck_barcode_file",
    )

    if "run_mode" in main_variables:
        parser.add_argument(
            "--run-mode",
            type=str,
            nargs="+",
            help="run_mode names for this sample.\n"
            + "the sample will be processed using the provided run_modes.\n"
            + "for merged samples, if left empty, the run_modes of the \n"
            + "merged (input) samples will be intersected.\n",
            dest="run_mode",
        )
        parser.add_argument(
            "--run_mode",
            type=str,
            nargs="+",
            help=argparse.SUPPRESS,
            dest="run_mode",
        )

    return parser


def get_sample_extra_info_parser():
    logger.debug("get_sample_extra_info_parser() called")
    parser = argparse.ArgumentParser(allow_abbrev=False, add_help=False)

    parser.add_argument(
        "--investigator",
        type=str,
        help="add the name of the investigator(s) responsible for this sample",
    )

    parser.add_argument("--experiment", type=str, help="description of the experiment")

    import datetime

    parser.add_argument(
        "--sequencing-date",
        type=datetime.date.fromisoformat,
        help="sequencing date of the sample. format: YYYY-MM-DD",
        dest="sequencing_date",
    )
    parser.add_argument(
        "--sequencing_date",
        type=datetime.date.fromisoformat,
        help=argparse.SUPPRESS,
        dest="sequencing_date",
    )

    return parser


def get_data_parser(reads_required=False):
    """
    Returns a parser which contain extra arguments for a given sample.
    The returned parser will contain the --R1, --R2, --longreads,
    --longread-signature, --barcode_flavor, --species, --puck,  --puck_barcode_file_id,
    --puck_barcode_file, --investigator, --experiment, --sequencing_date,
    --run_mode arguments.

    :param species_required: if true, the --species argument will be required
        during parsing.
    :param reads_required: if true, --R1, --R2, and --longreads arguments will be
        required during parsing.
    """
    parser = argparse.ArgumentParser(allow_abbrev=False, add_help=False)
    logger.debug("get_data_parser() called")
    parser.add_argument(
        "--R1",
        type=str,
        help=".fastq.gz file path to R1 reads",
        nargs="+",
        required=reads_required,
    )

    parser.add_argument(
        "--R2",
        type=str,
        help=".fastq.gz file path to R2 reads",
        nargs="+",
        required=reads_required,
    )

    parser.add_argument(
        "--reads",
        type=str,
        default="None",
        help="path to a CSV file listing reads for the sample (can be used for pooling reads and also for aggregating bulk samples into one big analysis with different cell barcodes for each sample)",
        required=reads_required,
    )

    parser.add_argument(
        "--dge",
        type=str,
        help="Path to dge matrix. spacemake can also handle already processed"
        + " digital expression data",
        required=reads_required,
    )

    parser.add_argument(
        "--longreads",
        type=str,
        help="fastq(.gz)|fq(.gz)|bam file path to pacbio long reads for library debugging",
        required=reads_required,
    )

    parser.add_argument(
        "--longread-signature",
        type=str,
        help="identify the expected longread signature (see longread.yaml)",
    )

    return parser


def get_set_remove_variable_subparsers(
    parent_parser, variable_name, func, prepend="", allow_multiple=False
):
    """get_set_remove_variable_subparsers.

    :param parent_parser:
    :param variable_name:
    :param func:
    :param prepend:
    :param allow_multiple:
    """
    if allow_multiple:
        nargs = "+"
    else:
        nargs = None

    logger.debug(f"get_set_remove_variable_subparsers(varname={variable_name}) called")

    def get_action_parser(action):
        """Create and return a parser for a specific action.

        :param action: The action (e.g., "set" or "remove")
        """
        # Convert `action` and `variable_name` to kebab-case for subcommand
        kebab_action_variable = f"{action}-{variable_name.replace('_', '-')}"
        snake_action_variable = f"{action}_{variable_name}"

        # Add the parser for the action
        action_parser = parent_parser.add_parser(
            kebab_action_variable,
            parents=[get_project_sample_parser(allow_multiple=True)],
            description=f"{action} {variable_name} for several projects/samples",
            help=f"{action} {variable_name} for several projects/samples",
        )

        # Add the kebab-case and legacy snake_case arguments
        action_parser.add_argument(
            f"--{prepend}{variable_name.replace('_', '-')}",
            f"--{prepend}{variable_name}",  # Legacy argument
            type=str,
            required=True,
            nargs=nargs,
            help=f"Specify the {variable_name} to {action}",
            dest=variable_name,  # Unified destination
        )

        action_parser.set_defaults(func=func, variable=variable_name, action=action)

        return action_parser

    set_parser = get_action_parser("set")

    if allow_multiple:
        set_parser.add_argument("--keep-old", action="store_true")

        remove_parser = get_action_parser("remove")


def get_action_sample_parser(parent_parser, action, func):
    """get_action_sample_parser.

    :param parent_parser:
    :param action:
    :param func:
    """
    logger.debug(f"get_action_sample_parser(action={action}) called")

    if action not in ["add", "update", "delete", "merge"]:
        raise ValueError(f"Invalid action: {action}")

    if action == "merge":
        parser_name = "merge-samples"
        msg = "merge samples"
        parents = [
            get_project_sample_parser(
                prepend="merged-", help_extra="of the newly created merged sample"
            ),
            get_project_sample_parser(
                allow_multiple=True, help_extra="of the samples to be merged"
            ),
        ]
    else:
        parser_name = f"{action}-sample"
        msg = f"{action} a sample"
        parents = [get_project_sample_parser()]

    # Add additional arguments based on action
    if action == "add":
        # add arguments for species, run_mode, barcode_flavor and puck
        parents.append(
            get_sample_main_variables_parser(
                defaults=True,
                species_required=True,
            )
        )
        # add arguments for R1/R1, dge, longread
        parents.append(get_data_parser())
        # add arguments for extra sample info
        parents.append(get_sample_extra_info_parser())
    elif action == "update":
        # add main variables parser
        parents.append(get_sample_main_variables_parser())
        # add arguments for R1/R1, dge, longread
        parents.append(get_data_parser())
        # add possibility to add extra info
        parents.append(get_sample_extra_info_parser())
    elif action == "merge":
        # add main variables parser
        # when merging, only let the user overwrite puck and run_mode
        parents.append(
            get_sample_main_variables_parser(
                main_variables=["run_mode", "puck"],
            )
        )

        # add possibility to add extra info
        parents.append(get_sample_extra_info_parser())

    # Add the parser with kebab-case name and backward-compatible snake_case alias
    sample_parser = parent_parser.add_parser(
        parser_name,
        description=msg,
        help=msg,
        parents=parents,
        aliases=[f"{action}_sample"],
    )

    sample_parser.set_defaults(func=func, action=action)


def setup_project_parser(parent_parser_subparsers):
    """setup_project_parser.

    :param attach_to:
    """
    parser = parent_parser_subparsers.add_parser(
        "projects",
        help="manage projects and samples",
        description="Using one of the subcommands specified below, it is possible to"
        + " add/update/remove projects and their settings",
    )
    subparsers = parser.add_subparsers()

    help_desc = {
        "add_sample_sheet": "add projects and samples from Illumina sample sheet",
        "add_samples_from_yaml": "add several samples at once from a .yaml file",
        "merge_samples": "merge several samples into one. "
        + "samples need to have the same species",
        "list": "list several project(s) and sample(s) and their settings",
    }

    # ADD/UPDATE/DELETE/MERGE sample(s)
    # get parsers for adding, deleting and updating a sample
    for action in ["add", "update", "delete"]:
        get_action_sample_parser(
            subparsers,
            action,
            # attach the function to the parser
            func=add_update_delete_sample_cmdline,
        )

    # get parser for merging
    get_action_sample_parser(subparsers, "merge", func=merge_samples_cmdline)

    # LIST PROJECTS
    # always show these variables
    always_show = [
        "species",
        "barcode_flavor",
        "run_mode",
        "map_strategy",
        "puck",
        "puck_barcode_file_id",
    ]
    import spacemake.project_df as spdf

    remaining_options = [
        x
        for x in spdf.ProjectDF.project_df_default_values.keys()
        if x not in always_show
    ]
    list_projects = subparsers.add_parser(
        "list",
        description=help_desc["list"],
        help=help_desc["list"],
        parents=[
            get_project_sample_parser(
                allow_multiple=True,
                help_extra="subset the data. If none provided, all will be displayed",
            )
        ],
    )
    list_projects.add_argument(
        "--variables",
        help="which extra variables to display per sample? "
        + f"{always_show} will always be shown.",
        choices=remaining_options,
        default=[],
        nargs="*",
    )
    list_projects.set_defaults(func=list_projects_cmdline, always_show=always_show)

    # ADD SAMPLE SHEET
    sample_add_sample_sheet = subparsers.add_parser(
        "add-sample-sheet",
        description=help_desc["add_sample_sheet"],
        help=help_desc["add_sample_sheet"],
        parents=[get_add_sample_sheet_parser()],
    )
    sample_add_sample_sheet_legacy = subparsers.add_parser(
        "add_sample_sheet",
        parents=[get_add_sample_sheet_parser()],
    )
    sample_add_sample_sheet.set_defaults(func=add_sample_sheet_cmdline)
    sample_add_sample_sheet_legacy.set_defaults(func=add_sample_sheet_cmdline)

    # ADD SAMPLES FROM YAML
    sample_add_samples_yaml = subparsers.add_parser(
        "add-samples-from-yaml",
        description=help_desc["add_samples_from_yaml"],
        help=help_desc["add_samples_from_yaml"],
    )
    sample_add_samples_yaml_legacy = subparsers.add_parser(
        "add_samples_from_yaml",
    )
    sample_add_samples_yaml.add_argument(
        "--samples_yaml",
        type=str,
        required=True,
        help="path to the .yaml file containing sample info",
    )
    sample_add_samples_yaml_legacy.add_argument(
        "--samples_yaml",
        type=str,
        required=True,
        help="path to the .yaml file containing sample info",
    )
    sample_add_samples_yaml.set_defaults(func=add_samples_from_yaml_cmdline)
    sample_add_samples_yaml_legacy.set_defaults(func=add_samples_from_yaml_cmdline)

    # get set/remove parser for each main variable
    # this will add parser for:
    # setting/removing run_mode
    # setting species
    # setting puck
    # setting barcode_flavor
    # NOTE: the ConfigFile.main_variable_sg2type keys determine
    # which commandline args will be defined for the parser.
    # this is a little un-intuitive...
    # TODO: cleaner factory functions for the commandline-parsers
    for main_var_sg, main_var_type in ConfigFile.main_variable_sg2type.items():
        allow_multiple = False
        if isinstance(main_var_type, str) and main_var_type.endswith("_list"):
            allow_multiple = True

        get_set_remove_variable_subparsers(
            subparsers,
            variable_name=main_var_sg,
            func=set_remove_variable_cmdline,
            allow_multiple=allow_multiple,
        )
    return parser


def get_run_parser():
    """get_run_parser."""
    parser = argparse.ArgumentParser(allow_abbrev=False, add_help=False)

    parser.add_argument(
        "--cores", default=1, type=int, help="number of cores to be used in total"
    )
    parser.add_argument(
        "--dryrun",
        "-n",
        action="store_true",
        help="invokes a dry snakemake run, printing only commands",
    )
    parser.add_argument(
        "--debug",
        default="",
        help=f"comma-separated list of logging-domains for which you want DEBUG output",
    )
    parser.add_argument(
        "--rerun-incomplete",
        "--ri",
        action="store_true",
        help="forces snakemake to rerun incompletely generated files",
    )
    parser.add_argument(
        "--keep-going",
        action="store_true",
        help="if a job fails, keep executing independent jobs",
    )
    parser.add_argument(
        "--printshellcmds",
        "-p",
        action="store_true",
        help="print shell commands for each rule, if exist",
    )
    parser.add_argument(
        "--debug-smk",
        action="store_true",
        help="activate extended snakemake debug output on DAG construction and resons for running rules etc...",
    )
    parser.add_argument(
        "--touch",
        "-t",
        action="store_true",
        help="rather than running the rules, just touch each file",
    )

    parser.add_argument(
        "--with-fastqc",
        "-wfqc",
        action="store_true",
        help="Run also fastqc as part of the spacemake run",
        dest="with_fastqc",
    )
    parser.add_argument(
        "--with_fastqc",
        action="store_true",
        help=argparse.SUPPRESS,
        dest="with_fastqc",
    )

    return parser


def setup_init_parser(parent_parser_subparsers):
    """setup_init_parser.

    :param parent_parser_subparsers:
    """
    parser_init = parent_parser_subparsers.add_parser(
        "init",
        help="initialise spacemake: create config files, download genomes and annotations",
    )

    parser_init.add_argument(
        "--root-dir",
        default="",
        help="where to output the results of the spacemake run. defaults to .",
        dest="root_dir",
    )
    parser_init.add_argument(
        "--root_dir",
        default="",
        help=argparse.SUPPRESS,
        dest="root_dir",
    )

    parser_init.add_argument(
        "--temp-dir",
        default="/tmp",
        help="temporary directory used when running spacemake. defaults to /tmp",
        dest="temp_dir",
    )
    parser_init.add_argument(
        "--temp_dir",
        default="/tmp",
        help=argparse.SUPPRESS,
        dest="temp_dir",
    )

    parser_init.add_argument(
        "--download-species",
        default=False,
        help="if set, upon initialisation, spacemake will download the mouse and human genome and index",
        action="store_true",
        dest="download_species",
    )
    parser_init.add_argument(
        "--download_species",
        default=False,
        help=argparse.SUPPRESS,
        action="store_true",
        dest="download_species",
    )

    # This enables a pre-configured container
    DST_from_env = os.environ.get("DROPSEQ_TOOLS", "")
    if DST_from_env:
        parser_init.add_argument(
            "--dropseq-tools",
            help="absolute path to dropseq_tools directory",
            required=False,
            default=DST_from_env,
            dest="dropseq_tools",
        )
    else:
        parser_init.add_argument(
            "--dropseq-tools",
            help="absolute path to dropseq_tools directory",
            required=True,
            dest="dropseq_tools",
        )
    parser_init.add_argument(
        "--dropseq_tools",
        help=argparse.SUPPRESS,
        dest="dropseq_tools",
    )

    parser_init.add_argument(
        "--species-directory",
        help="absolute path to directory containing species information (sequences, annotation, and indices)",
        required=False,
        dest="species_directory",
    )

    parser_init.set_defaults(func=spacemake_init)

    return parser_init


def setup_run_parser(parent_parser_subparsers):
    """setup_run_parser.

    :param parent_parser_subparsers:
    """
    parser_run = parent_parser_subparsers.add_parser(
        "run", help="run spacemake", parents=[get_run_parser()]
    )
    parser_run.set_defaults(func=spacemake_run)

    # create subparsers
    parser_run_subparsers = parser_run.add_subparsers()

    downsampling_parser = parser_run_subparsers.add_parser(
        "downsample",
        help="run downsampling analysis for a list of projects/samples",
        parents=[get_project_sample_parser(allow_multiple=True), get_run_parser()],
    )
    downsampling_parser.set_defaults(downsample=True, func=spacemake_run)

    # parser_novosparc = novosparc_spacemake_parser(parser_run_subparsers)
    # parser_novosparc.set_defaults(novosparc_reconstruct=True,
    #    func=lambda args: spacemake_run(pdf, args))

    return parser_run


def setup_migrate_parser(parent_parser_subparsers):
    """setup_migrate_parser

    :param parent_parser_subparsers
    """
    parser_migrate = parent_parser_subparsers.add_parser(
        "migrate", help="migrate spacemake to the latest version"
    )

    parser_migrate.add_argument(
        "--project-id",
        help="project-id to migrate",
        type=str,
        dest="project_id",
    )
    parser_migrate.add_argument(
        "--sample-id",
        help="sample-id to migrate",
        type=str,
        dest="sample_id",
    )
    parser_migrate.add_argument(
        "--cores",
        default="4",
        help="number of cores to use",
        type=int,
    )
    parser_migrate.add_argument(
        "--all",
        action="store_true",
        help="if set, spacemake migrates all samples",
    )
    parser_migrate.add_argument(
        "--dryrun",
        "-n",
        action="store_true",
        help="invokes a dry snakemake run, printing only commands",
    )
    parser_migrate.add_argument(
        "--debug",
        default="",
        help=f"comma-separated list of logging-domains for which you want DEBUG output",
    )
    parser_migrate.add_argument(
        "--rerun-incomplete",
        "--ri",
        action="store_true",
        help="forces snakemake to rerun incompletely generated files",
    )
    parser_migrate.add_argument(
        "--keep-going",
        action="store_true",
        help="if a job fails, keep executing independent jobs",
    )
    parser_migrate.add_argument(
        "--printshellcmds",
        "-p",
        action="store_true",
        help="print shell commands for each rule, if exist",
    )
    parser_migrate.add_argument(
        "--touch",
        "-t",
        action="store_true",
        help="rather than running the rules, just touch each file",
    )

    parser_migrate.set_defaults(func=spacemake_migrate)

    return parser_migrate


#####################################################
# actual command-line functions, used as call-backs #
#####################################################


@message_aggregation(logger_name)
def spacemake_init(args):
    """spacemake_init.

    :param args:
    """
    import snakemake
    import yaml
    from spacemake.snakemake.variables import config_path

    if os.path.isfile(config_path):
        msg = "spacemake has already been initiated in this directory.\n"
        msg += "use other commands to run and analyse samples."
        logger.info(msg)
        return 0
    initial_config = os.path.join(os.path.dirname(__file__), "data/config/config.yaml")

    # initialise config file
    cf = ConfigFile.from_yaml(initial_config)
    # update the file path
    cf.set_file_path(config_path)

    # add variables from args
    cf.variables["root_dir"] = args["root_dir"]
    cf.variables["temp_dir"] = args["temp_dir"]

    cf.variables["external_bin"] = {"dropseq_tools": args["dropseq_tools"]}

    cf.variables["microscopy_out"] = args.get("microscopy_out", "")

    if args["download_species"]:
        species_data_config_file = os.path.join(
            os.path.dirname(__file__), "data/config/species_data_url.yaml"
        )
        species_data_config = yaml.load(
            open(species_data_config_file), Loader=yaml.FullLoader
        )

        # add keys as species to config
        species = list(species_data_config.keys())
        snakemake_config = dict(
            root_dir="",
            species=species,
            # define the pattern
            annotation_file_pattern="species_data/{species}/{species}_{data_type}.gtf",
            genome_file_pattern="species_data/{species}/{species}_{data_type}.fa",
        )

        # the to be saved file paths
        species_info = {}
        species_files_exist = []

        for sp in species:
            if not sp in species_info.keys():
                species_info[sp] = {}
            for data_type in ["genome", "annotation"]:
                # save download url to be passed to snakemake
                snakemake_config[sp + "_" + data_type + "_url"] = species_data_config[
                    sp
                ][data_type]
                # save file path of the result
                species_file = snakemake_config[data_type + "_file_pattern"].format(
                    species=sp, data_type=data_type
                )
                species_info[sp][data_type] = species_file

                # add bool if file exists
                species_files_exist.append(os.path.isfile(species_file))

        # if any of the files are missing
        if not all(species_files_exist):
            # get the snakefile for downloading the .gtf and .fa files
            snakefile = os.path.join(
                os.path.dirname(__file__), "snakemake/species_init.smk"
            )
            # run snakemake: download species data and place them in the right place
            snakemake.snakemake(
                snakefile, cores=1, config=snakemake_config, rerun_triggers=["mtime"]
            )

        for key, value in species_info.items():
            cf.add_variable(
                "species",
                key,
                reference="genome",
                sequence=value["genome"],
                annotation=value["annotation"],
            )

    if "species_directory" in args:
        # future code to create the species folder appropriately and symlink inside etc
        print("")

    # copy visium_puck_barcode_file
    dest_visium_path = "puck_data/visium_barcode_positions.csv"
    logger.info(f"Moving visium barcodes to {dest_visium_path}")
    os.makedirs(os.path.dirname(dest_visium_path), exist_ok=True)
    from shutil import copyfile

    copyfile(
        os.path.join(os.path.dirname(__file__), "data/visium_barcode_positions.csv"),
        dest_visium_path,
    )

    # copy openst_coordinate_system
    dest_puck_collection_path = "puck_data/openst_coordinate_system.csv"
    logger.info(
        f"Moving puck collection coordinate system to {dest_puck_collection_path}"
    )
    os.makedirs(os.path.dirname(dest_puck_collection_path), exist_ok=True)
    copyfile(
        os.path.join(
            os.path.dirname(__file__),
            "data/puck_collection/openst_coordinate_system.csv",
        ),
        dest_puck_collection_path,
    )

    # save
    cf.dump()


def collect_smk_options(args):
    import spacemake.snakemake.variables as var

    smk_options = dict(
        configfiles=[var.config_path],
        cores=args["cores"],
        dryrun=args["dryrun"],
        touch=args.get("touch", False),
        force_incomplete=args.get("rerun_incomplete", False),
        keepgoing=args.get("keep_going", False),
        printshellcmds=args.get("printshellcmds", False),
        rerun_triggers=["mtime"],  # needed for newer snakemake
    )
    if args.get("debug_smk", False):
        smk_options.update(
            dict(
                debug_dag=True,  # TODO: make these configureable
                verbose=True,
            )
        )

    return smk_options


@message_aggregation(logger_name)
def spacemake_run(args):
    """spacemake_run.

    :param args:
    """
    from spacemake.errors import SpacemakeError
    import snakemake

    import spacemake.snakemake.variables as var

    if not os.path.isfile(var.config_path):
        msg = "spacemake has not been initalised yet.\n"
        msg += "please run `spacemake init` to start a new project"

        raise SpacemakeError(msg)

    from spacemake.project_df import get_global_ProjectDF

    root = logging.getLogger("spacemake")
    # TODO: Why is args a dictionary and not argparse.Namespace ?
    if args["debug"]:
        # activate cmdline requested debug output for specific domains (comma-separated)
        for logger_name in args["debug"].split(","):
            if logger_name:
                root.info(f"setting domain {logger_name} to DEBUG")
                logging.getLogger(logger_name.replace("root", "")).setLevel(
                    logging.DEBUG
                )

    # Check if latest version spacemake is installed. If not: exit and ask for migration.
    # Current logic applies to versions 0.8x vs 0.9
    with open("config.yaml") as yamlfile:
        cf = yaml.safe_load(yamlfile.read())
    if "spacemake_version" not in cf:
        print(
            """
        ###############################################################
        #                                                             #
        #      WARNING: incompatible version detected                 #
        #                                                             #
        #   You are using a newer version of spacemake which is       #
        #   incompatible with your current project setup.             #
        #                                                             #
        #     Please migrate your projects using:                     #
        #      $ spacemake migrate                                    #
        #                                                             #
        #   For more info, see:                                       #
        #   https://spacemake.readthedocs.io/en/latest/migrate.html   #
        #                                                             #
        ###############################################################
        """
        )
        return

    pdf = get_global_ProjectDF()
    samples = []
    projects = []
    targets = ["run_analysis"]
    with_fastqc = args.get("with_fastqc", False)

    downsample = args.get("downsample", False)
    novosparc_reconstruct = args.get("novosparc_reconstruct", False)

    if downsample:
        targets = ["downsample"]
        samples = args.get("sample_id_list", [])
        projects = args.get("project_id_list", [])

    if novosparc_reconstruct:
        targets = ["novosparc"]
        import spacemake.smk as smk

        novosparc_variables = smk.get_novosparc_variables(pdf, args)
    else:
        novosparc_variables = {}

    config_variables = {
        "project_df": pdf.file_path,
        "samples": samples,
        "projects": projects,
        "with_fastqc": with_fastqc,
        "pwd": os.getcwd(),
        "log_debug": args["debug"],
    }

    # join config_variables and novosparc_variables
    # to flatten the diunpack(get_final_bam)e
    snakefile = os.path.join(os.path.dirname(__file__), "snakemake/main.smk")
    # run snakemake

    smk_options = collect_smk_options(args)

    preprocess_finished = snakemake.snakemake(
        snakefile,
        targets=["get_stats_prealigned_barcodes", "unload_genome_flag"],
        config=config_variables,
        **smk_options,
    )
    if preprocess_finished is False:
        raise SpacemakeError("an error occurred while snakemake() ran")

    # update valid pucks (above threshold) before continuing to downstream
    # this performs counting of matching barcodes after alignment
    if not args["dryrun"]:
        pdf.update_project_df_barcode_matches(prealigned=True)
        pdf.consolidate_pucks_merged_samples()
        pdf.dump()

    # whitelisting of barcodes
    preprocess_finished = snakemake.snakemake(
        snakefile,
        targets=["get_whitelist_barcodes"],
        config=config_variables,
        **smk_options,
    )

    if not args["dryrun"]:
        pdf.update_project_df_barcode_matches()
        pdf.dump()

    # run snakemake quantification and reports
    analysis_finished = snakemake.snakemake(
        snakefile, targets=targets, config=config_variables, **smk_options
    )

    if analysis_finished is False:
        raise SpacemakeError("an error occurred while snakemake() ran")

    # at the very end dump the project_data_frame
    if not args["dryrun"]:
        pdf.dump()


@message_aggregation(logger_name)
def add_update_delete_sample_cmdline(args):
    """add_update_delete_sample_cmdline.

    :param args:
    """
    action = args["action"]
    from spacemake.project_df import get_global_ProjectDF

    pdf = get_global_ProjectDF()
    # remove the action from args
    del args["action"]

    if action == "add" or action == "update":
        func = lambda **kwargs: pdf.add_update_sample(action=action, **kwargs)
    elif action == "delete":
        func = pdf.delete_sample

    sample = func(**args)
    pdf.dump()


@message_aggregation(logger_name)
def add_sample_sheet_cmdline(args):
    """add_sample_sheet_cmdline.

    :param args:
    """
    from spacemake.project_df import get_global_ProjectDF

    pdf = get_global_ProjectDF()

    pdf.add_sample_sheet(args["sample_sheet"], args["basecalls_dir"])

    pdf.dump()


@message_aggregation(logger_name)
def add_samples_from_yaml_cmdline(args):
    """add_samples_from_yaml_cmdline.

    :param args:
    """
    from spacemake.project_df import get_global_ProjectDF

    pdf = get_global_ProjectDF()
    pdf.add_samples_from_yaml(args["samples_yaml"])

    pdf.dump()


@message_aggregation(logger_name)
def set_remove_variable_cmdline(args):
    """set_remove_variable_cmdline.

    :param args:
    """
    variable_name = args["variable"]
    action = args["action"]

    variable_key = args[variable_name]
    from spacemake.project_df import get_global_ProjectDF

    pdf = get_global_ProjectDF()

    pdf.set_remove_variable(
        variable_name=variable_name,
        variable_key=variable_key,
        action=action,
        project_id_list=args["project_id_list"],
        sample_id_list=args["sample_id_list"],
        keep_old=args.get("keep_old", False),
    )

    pdf.dump()


@message_aggregation(logger_name)
def merge_samples_cmdline(args):
    """merge_samples_cmdline.

    :param args:
    """
    from spacemake.project_df import get_global_ProjectDF

    pdf = get_global_ProjectDF()
    pdf.merge_samples(**args)

    pdf.dump()


@message_aggregation(logger_name)
def list_projects_cmdline(args):
    """list_projects_cmdline.

    :param args:
    """
    from spacemake.project_df import get_global_ProjectDF

    pdf = get_global_ProjectDF()

    projects = args["project_id_list"]
    samples = args["sample_id_list"]
    variables = args["always_show"] + args["variables"]

    df = pdf.df
    logger = logging.getLogger(logger_name)

    if projects != [] or samples != []:
        df = df.query("project_id in @projects or sample_id in @samples")
        logger.inof(f"listing projects: {projects} and samples: {samples}")
    else:
        logger.info("listing all projects and samples")

    logger.info(f"variables used: {variables}")

    # print the table
    print(df.loc[:, variables].__str__())
    # logger.info()


@message_aggregation(logger_name)
def spacemake_migrate(args):
    """spacemake_migrate.

    :param args:
    """
    import snakemake
    from spacemake.errors import SpacemakeError
    import spacemake.snakemake.variables as var
    from spacemake.project_df import get_global_ProjectDF

    # check if it's a legit spacemake folder
    if not os.path.isfile(var.config_path):
        msg = "spacemake has not been initalised yet in this folder.\n"
        msg += "please run `spacemake init` to start a new project"
        raise SpacemakeError(msg)

    pdf = get_global_ProjectDF()

    def build_sample_info(project_id, sample_id):
        """Helper function to extract project_folder and target."""
        project_folder = os.path.join(
            f"projects/{project_id}/processed_data/{sample_id}/illumina/complete_data"
        )
        target = os.path.join(project_folder, "final.polyA_adapter_trimmed.cram")
        return project_id, sample_id, project_folder, target

    # check if either the (--project_id, --sample_id) OR the --all flag were provided
    if args.get("all"):
        if args.get("project_id") or args.get("sample_id"):
            raise ValueError(
                "Do not provide --project-id or --sample-id when using --all."
            )
        sample_records = pdf.df.index.tolist()
    else:
        if not args.get("project_id") or not args.get("sample_id"):
            raise ValueError(
                "Both --project-id and --sample-id are required unless using --all."
            )
        sample_records = [(args["project_id"], args["sample_id"])]

    # build per-sample info
    sample_infos = [build_sample_info(p, s) for p, s in sample_records]
    projects, samples, project_folders, targets = map(list, zip(*sample_infos))

    # build config
    config_variables = {
        "project_df": pdf.file_path,
        "pwd": os.getcwd(),
        "projects": projects,
        "samples": samples,
        "project_folders": project_folders,
        "targets": targets,
        # "project_sample_folder_triplets": list(zip(projects, samples, project_folders)),
    }

    # get the migration snakefile
    snakefile = os.path.join(os.path.dirname(__file__), "snakemake/migrate.smk")

    # run snakemake
    migration_finished = snakemake.snakemake(
        snakefile,
        config=config_variables,
        **collect_smk_options(args),
    )
    if migration_finished is False:
        raise SpacemakeError("an error occurred while snakemake() ran")

    return


def make_main_parser():
    #################
    # DEFINE PARSER #
    #################
    # we add main parser to a dictionary
    # so that we can later call the help function
    # based on the sub-command. this is to print the
    # -h (help) functionality if no parameters are provided,
    # rather than printing an error.
    # import spacemake.smk as smk

    parser_main = argparse.ArgumentParser(
        allow_abbrev=False,
        description="Spacemake: processing and analysis of large-scale spatial transcriptomics data",
    )

    parser_main.add_argument("--version", action="store_true")

    parser_main_subparsers = parser_main.add_subparsers(
        help="sub-command help", dest="subcommand"
    )

    parser_run = None
    parser_projects = None
    parser_config = None
    parser_init = None
    parser_migrate = None
    parser_spatial = None

    ##################
    # SPACEMAKE INIT #
    ##################
    parser_init = setup_init_parser(parser_main_subparsers)

    ####################
    # SPACEMAKE CONFIG #
    ####################
    ## spacemake_config args
    from spacemake.config import setup_config_parser
    from spacemake.snakemake.variables import config_path

    if os.path.isfile(config_path):
        # save config file
        parser_config = setup_config_parser(parser_main_subparsers)

        ############################
        # SPACEMAKE PROJECT/SAMPLE #
        ############################
        from spacemake.cmdline import setup_project_parser

        parser_projects = setup_project_parser(parser_main_subparsers)

        #################
        # SPACEMAKE RUN #
        #################
        parser_run = setup_run_parser(parser_main_subparsers)

        #####################
        # SPACEMAKE MIGRATE #
        #####################
        parser_migrate = setup_migrate_parser(parser_main_subparsers)

        #####################
        # SPACEMAKE SPATIAL #
        #####################
        from spacemake.spatial.cmdline import setup_spatial_parser

        parser_spatial = setup_spatial_parser(parser_main_subparsers)

    parser_dict = {
        "init": parser_init,
        "config": parser_config,
        "projects": parser_projects,
        "run": parser_run,
        "migrate": parser_migrate,
        "main": parser_main,
        "spatial": parser_spatial,
    }

    return parser_main, parser_dict


def cmdline():
    # import importlib.metadata
    """cmdline."""
    parser_main, parser_dict = make_main_parser()
    args = parser_main.parse_args()

    if args.version and args.subcommand is None:
        import spacemake.contrib

        print(spacemake.contrib.__version__)
        return 0
    else:
        del args.version

    from spacemake.util import setup_logging

    setup_logging(args, name="spacemake", rename_process=False, sample="cmdline()")
    # get the function to be run
    if "func" in args:
        func = args.func
    # else print help
    else:
        if args.subcommand is not None:
            parser_dict[args.subcommand].print_help()
        else:
            parser_dict["main"].print_help()
        return 0

    # get the args and delete the func key, get only set values
    args = {key: value for key, value in vars(args).items() if value is not None}
    args.pop("func", None)
    # pop also main,
    args.pop("subcommand", None)

    return func(args)


if __name__ == "__main__":
    cmdline()
